from __future__ import annotations

from types import SimpleNamespace
from typing import Mapping, Sequence
import sys

import pytest

from open_data_contract_standard.model import OpenDataContractStandard, Server  # type: ignore

from dc43_service_backends.config import (
    ContractStoreConfig,
    DataProductStoreConfig,
    GovernanceStoreConfig,
    ServiceBackendsConfig,
    UnityCatalogConfig,
)
from dc43_service_backends.governance.backend.local import LocalGovernanceServiceBackend
from dc43_service_backends.governance.unity_catalog import (
    UnityCatalogLinker,
    contract_servers_table_resolver,
    build_link_hooks,
    build_linker_from_config,
    prefix_table_resolver,
    sql_table_property_updater,
    sql_table_tag_updater,
)
from dc43_service_backends.governance.bootstrap import (
    DEFAULT_LINK_HOOK_BUILDER_SPECS,
    LinkHookContext,
    build_dataset_contract_link_hooks,
    load_link_hook_builder,
)


def test_linker_updates_table() -> None:
    updates: list[tuple[str, Mapping[str, str]]] = []

    def _update(table_name: str, properties: Mapping[str, str]) -> None:
        updates.append((table_name, dict(properties)))

    linker = UnityCatalogLinker(
        apply_table_properties=_update,
        static_properties={"dc43.catalog_synced": "true"},
    )

    linker.link_dataset_contract(
        dataset_id="table:governed.analytics.orders",
        dataset_version="42",
        contract_id="sales.orders",
        contract_version="0.1.0",
    )

    assert updates == [
        (
            "governed.analytics.orders",
            {
                "dc43.contract_id": "sales.orders",
                "dc43.contract_version": "0.1.0",
                "dc43.dataset_version": "42",
                "dc43.catalog_synced": "true",
            },
        )
    ]


def test_linker_uses_contract_servers_before_dataset_id() -> None:
    updates: list[str] = []

    def _update(table_name: str, properties: Mapping[str, str]) -> None:
        updates.append(table_name)

    contract = OpenDataContractStandard(
        id="sales.orders",
        version="0.1.0",
        kind="DatasetContract",
        apiVersion="3.0.2",
        servers=[
            Server(server="unity", type="catalog", catalog="dev", schema="bronze", dataset="orders"),
            Server(server="unity", type="catalog", catalog="dev", schema="bronze", dataset="orders"),
        ],
    )

    linker = UnityCatalogLinker(
        apply_table_properties=_update,
        contract_loader=lambda *_: contract,
        contract_table_resolver=contract_servers_table_resolver,
        table_resolver=prefix_table_resolver("table__"),
    )

    linker.link_dataset_contract(
        dataset_id="ignored.dataset",  # does not match the prefix
        dataset_version="7",
        contract_id="sales.orders",
        contract_version="0.1.0",
    )

    assert updates == ["dev.bronze.orders"]


def test_linker_falls_back_to_dataset_identifier_when_contract_missing() -> None:
    updates: list[str] = []

    def _update(table_name: str, properties: Mapping[str, str]) -> None:
        updates.append(table_name)

    linker = UnityCatalogLinker(
        apply_table_properties=_update,
        contract_loader=lambda *_: None,
        table_resolver=prefix_table_resolver("table:"),
    )

    linker.link_dataset_contract(
        dataset_id="table:dev.analytics.orders",
        dataset_version="9",
        contract_id="sales.orders",
        contract_version="0.2.0",
    )

    assert updates == ["dev.analytics.orders"]


def test_contract_servers_table_resolver_handles_partial_servers() -> None:
    contract = OpenDataContractStandard(
        id="sales.orders",
        version="0.1.0",
        kind="DatasetContract",
        apiVersion="3.0.2",
        servers=[
            Server(server="unity", type="catalog", catalog="dev", schema="bronze", dataset="orders"),
            Server(server="unity", type="catalog", schema="bronze", dataset="orders_rejects"),
            Server(server="raw", type="filesystem", path="dbfs:/raw/orders"),
        ],
    )

    tables = contract_servers_table_resolver(contract)

    assert tables == ("dev.bronze.orders", "bronze.orders_rejects")


def test_linker_skips_non_matching_datasets() -> None:
    updates: list[tuple[str, Mapping[str, str]]] = []

    def _update(table_name: str, properties: Mapping[str, str]) -> None:
        updates.append((table_name, dict(properties)))

    linker = UnityCatalogLinker(
        apply_table_properties=_update,
        table_resolver=prefix_table_resolver("table:"),
    )

    linker.link_dataset_contract(
        dataset_id="path:dbfs:/tmp/orders",
        dataset_version="7",
        contract_id="sales.orders",
        contract_version="0.1.0",
    )

    assert updates == []


def test_linker_skips_reserved_tables() -> None:
    updates: list[tuple[str, Mapping[str, str]]] = []

    def _update(table_name: str, properties: Mapping[str, str]) -> None:
        updates.append((table_name, dict(properties)))

    linker = UnityCatalogLinker(
        apply_table_properties=_update,
        skip_tables=frozenset({"dev.catalog.contracts"}),
    )

    with pytest.warns(RuntimeWarning):
        linker.link_dataset_contract(
            dataset_id="table:dev.catalog.contracts",
            dataset_version="1",
            contract_id="sales.orders",
            contract_version="0.1.0",
        )

    assert updates == []


def test_linker_updates_tags_and_unsets() -> None:
    calls: list[tuple[str, Mapping[str, str], tuple[str, ...]]] = []

    def _tags(table_name: str, tags: Mapping[str, str], unset_tags: Sequence[str] = ()) -> None:
        calls.append((table_name, dict(tags), tuple(unset_tags)))

    linker = UnityCatalogLinker(
        apply_table_tags=_tags,
        static_tags={"dc43.catalog_synced": "true"},
    )

    linker.link_dataset_contract(
        dataset_id="table:governed.analytics.orders",
        dataset_version="42",
        contract_id="sales.orders",
        contract_version="0.1.0",
    )

    def _metadata(*_: str) -> Mapping[str, object]:
        return {"dc43.contract_id": None}

    linker.metadata_provider = _metadata
    linker.static_tags = {}

    linker.link_dataset_contract(
        dataset_id="table:governed.analytics.orders",
        dataset_version="",
        contract_id="",
        contract_version="",
    )

    assert calls == [
        (
            "governed.analytics.orders",
            {
                "dc43_contract_id": "sales.orders",
                "dc43_contract_version": "0.1.0",
                "dc43_dataset_version": "42",
                "dc43_catalog_synced": "true",
            },
            (),
        ),
        ("governed.analytics.orders", {}, ("dc43_contract_id",)),
    ]


def test_linker_ignores_reserved_property_keys() -> None:
    updates: list[tuple[str, Mapping[str, str]]] = []

    def _update(table_name: str, properties: Mapping[str, str]) -> None:
        updates.append((table_name, dict(properties)))

    linker = UnityCatalogLinker(
        apply_table_properties=_update,
        static_properties={"owner": "team"},
    )

    with pytest.warns(RuntimeWarning):
        linker.link_dataset_contract(
            dataset_id="table:governed.analytics.orders",
            dataset_version="42",
            contract_id="sales.orders",
            contract_version="0.1.0",
        )

    assert updates == [
        (
            "governed.analytics.orders",
            {
                "dc43.contract_id": "sales.orders",
                "dc43.contract_version": "0.1.0",
                "dc43.dataset_version": "42",
            },
        )
    ]


def test_linker_converts_invalid_tag_names() -> None:
    calls: list[tuple[str, Mapping[str, str], tuple[str, ...]]] = []

    def _tags(table_name: str, tags: Mapping[str, str], unset_tags: Sequence[str] = ()) -> None:
        calls.append((table_name, dict(tags), tuple(unset_tags)))

    def _metadata(*_: str) -> Mapping[str, object]:
        return {
            "dc43.contract:id": "sales.orders",
            "dc43.contract_version": "0.1.0",
            "dc43.dataset_version": "42",
        }

    linker = UnityCatalogLinker(apply_table_tags=_tags)
    linker.metadata_provider = _metadata

    with pytest.warns(RuntimeWarning):
        linker.link_dataset_contract(
            dataset_id="table:governed.analytics.orders",
            dataset_version="42",
            contract_id="sales.orders",
            contract_version="0.1.0",
        )

    assert calls == [
        (
            "governed.analytics.orders",
            {"dc43_contract_id": "sales.orders", "dc43_contract_version": "0.1.0", "dc43_dataset_version": "42"},
            (),
        )
    ]


def test_linker_handles_property_errors() -> None:
    failures: list[str] = []

    def _update(table_name: str, properties: Mapping[str, str]) -> None:
        failures.append(table_name)
        raise RuntimeError("permission denied")

    linker = UnityCatalogLinker(apply_table_properties=_update)

    with pytest.warns(RuntimeWarning):
        linker.link_dataset_contract(
            dataset_id="table:governed.analytics.orders",
            dataset_version="42",
            contract_id="sales.orders",
            contract_version="0.1.0",
        )

    assert failures == ["governed.analytics.orders"]


class _SQLConnection:
    def __init__(self, engine: "_SQLEngine") -> None:
        self._engine = engine

    def execute(self, statement) -> None:  # pragma: no cover - exercised via updater
        self._engine.statements.append(str(statement))


class _SQLEngine:
    def __init__(self) -> None:
        self.statements: list[str] = []

    def begin(self):  # pragma: no cover - exercised via updater
        engine = self

        class _Transaction:
            def __enter__(self):
                return _SQLConnection(engine)

            def __exit__(self, exc_type, exc, tb) -> None:
                return None

        return _Transaction()


def test_sql_table_property_updater_executes_statements() -> None:
    engine = _SQLEngine()
    updater = sql_table_property_updater(engine)

    updater("dev.catalog.orders", {"dc43.contract_id": "sales.orders", "env": "dev"})

    assert engine.statements == [
        "ALTER TABLE `dev`.`catalog`.`orders` SET TBLPROPERTIES ('dc43.contract_id'='sales.orders', 'env'='dev')"
    ]


def test_sql_table_tag_updater_executes_statements() -> None:
    engine = _SQLEngine()
    updater = sql_table_tag_updater(engine)

    updater("dev.catalog.orders", {"owner": "team"}, ("outdated",))

    assert engine.statements == [
        "ALTER TABLE `dev`.`catalog`.`orders` UNSET TAGS ('outdated')",
        "ALTER TABLE `dev`.`catalog`.`orders` SET TAGS ('owner'='team')",
    ]


class _ContractClient:
    def __init__(self) -> None:
        self.links: list[tuple[str, str, str, str]] = []

    def link_dataset_contract(
        self,
        *,
        dataset_id: str,
        dataset_version: str,
        contract_id: str,
        contract_version: str,
    ) -> None:
        self.links.append((dataset_id, dataset_version, contract_id, contract_version))


class _DQClient:
    pass


def test_local_backend_still_tags_with_remote_contract_client() -> None:
    updates: list[tuple[str, Mapping[str, str]]] = []

    def _update(table_name: str, properties: Mapping[str, str]) -> None:
        updates.append((table_name, dict(properties)))

    contract_client = _ContractClient()
    backend = LocalGovernanceServiceBackend(
        contract_client=contract_client,
        dq_client=_DQClient(),
        link_hooks=[UnityCatalogLinker(apply_table_properties=_update).link_dataset_contract],
    )

    backend.link_dataset_contract(
        dataset_id="table:governed.analytics.orders",
        dataset_version="1",
        contract_id="sales.orders",
        contract_version="0.1.0",
    )

    assert contract_client.links == [
        ("table:governed.analytics.orders", "1", "sales.orders", "0.1.0")
    ]
    assert updates == [
        (
            "governed.analytics.orders",
            {
                "dc43.contract_id": "sales.orders",
                "dc43.contract_version": "0.1.0",
                "dc43.dataset_version": "1",
            },
        )
    ]


def test_build_linker_from_config_prefers_sql_engine(monkeypatch: pytest.MonkeyPatch) -> None:
    engine = _SQLEngine()
    dsns: list[str] = []

    def _engine_builder(dsn: str) -> _SQLEngine:
        dsns.append(dsn)
        return engine

    config = UnityCatalogConfig(
        enabled=True,
        sql_dsn="databricks://token:abc@adb.example.com?http_path=/sql/warehouses/demo",
        dataset_prefix="table:",
    )

    linker = build_linker_from_config(
        config,
        engine_builder=_engine_builder,
    )

    assert isinstance(linker, UnityCatalogLinker)
    assert dsns == ["databricks://token:abc@adb.example.com?http_path=/sql/warehouses/demo"]

    linker.link_dataset_contract(
        dataset_id="table:dev.catalog.orders",
        dataset_version="1",
        contract_id="sales.orders",
        contract_version="0.3.0",
    )

    assert engine.statements == [
        "ALTER TABLE `dev`.`catalog`.`orders` SET TBLPROPERTIES ('dc43.contract_id'='sales.orders', 'dc43.contract_version'='0.3.0', 'dc43.dataset_version'='1')"
    ]


def test_build_linker_from_config_with_tags(monkeypatch: pytest.MonkeyPatch) -> None:
    engine = _SQLEngine()

    def _engine_builder(dsn: str) -> _SQLEngine:
        return engine

    config = UnityCatalogConfig(
        enabled=True,
        sql_dsn="databricks://token@example",  # property path
        tags_enabled=True,
        static_tags={"env": "dev"},
    )

    linker = build_linker_from_config(
        config,
        engine_builder=_engine_builder,
    )

    assert isinstance(linker, UnityCatalogLinker)

    linker.link_dataset_contract(
        dataset_id="table:dev.catalog.orders",
        dataset_version="2",
        contract_id="sales.orders",
        contract_version="0.4.0",
    )

    assert engine.statements == [
        "ALTER TABLE `dev`.`catalog`.`orders` SET TBLPROPERTIES ('dc43.contract_id'='sales.orders', 'dc43.contract_version'='0.4.0', 'dc43.dataset_version'='2')",
        "ALTER TABLE `dev`.`catalog`.`orders` SET TAGS ('dc43_contract_id'='sales.orders', 'dc43_contract_version'='0.4.0', 'dc43_dataset_version'='2', 'env'='dev')",
    ]


def test_build_link_hooks_injects_reserved_tables(monkeypatch: pytest.MonkeyPatch) -> None:
    captured: dict[str, tuple[str, ...]] = {}

    def _fake_builder(
        _: UnityCatalogConfig,
        context=None,
        engine_builder=sql_table_property_updater,  # type: ignore[assignment]
        *,
        skip_tables: Sequence[str] | None = None,
        **extra: object,
    ) -> UnityCatalogLinker:
        captured["skip"] = tuple(sorted(skip_tables or ()))

        class _Linker:
            def link_dataset_contract(self, **_: object) -> None:
                pass

        return _Linker()  # type: ignore[return-value]

    monkeypatch.setattr(
        "dc43_service_backends.governance.unity_catalog.build_linker_from_config",
        _fake_builder,
    )

    config = ServiceBackendsConfig(
        unity_catalog=UnityCatalogConfig(enabled=True, sql_dsn="sqlite:///tmp.db"),
        contract_store=ContractStoreConfig(table="dev.catalog.contracts"),
        data_product_store=DataProductStoreConfig(table="dev.catalog.products"),
        governance_store=GovernanceStoreConfig(
            table="dev.catalog.dq_status",
            status_table="dev.catalog.dq_status",
            activity_table="dev.catalog.dq_activity",
            link_table="dev.catalog.dq_links",
            metrics_table="dev.catalog.dq_metrics",
        ),
    )

    hooks = build_link_hooks(config)

    assert hooks is not None and len(hooks) == 1
    assert set(captured["skip"]) == {
        "dev.catalog.contracts",
        "dev.catalog.products",
        "dev.catalog.dq_status",
        "dev.catalog.dq_activity",
        "dev.catalog.dq_links",
        "dev.catalog.dq_metrics",
    }


def test_build_linker_from_config_warns_when_tag_dsn_missing(monkeypatch: pytest.MonkeyPatch) -> None:
    config = UnityCatalogConfig(enabled=True, tags_enabled=True)

    with pytest.warns(RuntimeWarning) as recorded:
        linker = build_linker_from_config(config)

    assert linker is None
    assert len(recorded) >= 2


def test_build_linker_from_config_warns_without_sql_dsn() -> None:
    config = UnityCatalogConfig(enabled=True)

    with pytest.warns(RuntimeWarning):
        linker = build_linker_from_config(config)

    assert linker is None


def test_build_linker_from_config_supports_tag_only_flow() -> None:
    tag_engine = _SQLEngine()

    def _engine_builder(dsn: str) -> _SQLEngine:
        assert dsn == "databricks://token:abc@adb.example.com?http_path=/sql/warehouses/tags"
        return tag_engine

    config = UnityCatalogConfig(
        enabled=True,
        tags_enabled=True,
        tags_sql_dsn="databricks://token:abc@adb.example.com?http_path=/sql/warehouses/tags",
    )

    with pytest.warns(RuntimeWarning):
        linker = build_linker_from_config(config, engine_builder=_engine_builder)

    assert isinstance(linker, UnityCatalogLinker)

    linker.link_dataset_contract(
        dataset_id="table:dev.catalog.orders",
        dataset_version="2",
        contract_id="sales.orders",
        contract_version="0.4.0",
    )

    assert tag_engine.statements == [
        "ALTER TABLE `dev`.`catalog`.`orders` SET TAGS ('dc43_contract_id'='sales.orders', 'dc43_contract_version'='0.4.0', 'dc43_dataset_version'='2')",
    ]


def test_build_link_hooks_passes_contract_loader(monkeypatch: pytest.MonkeyPatch) -> None:
    captured: dict[str, object] = {}

    class _ContractService:
        def get(self, contract_id: str, contract_version: str) -> None:
            captured["requested"] = (contract_id, contract_version)
            return None

    class _Linker:
        def link_dataset_contract(self, **_: object) -> None:
            pass

    def _fake_builder(
        config: UnityCatalogConfig,
        *,
        contract_loader=None,
        **_: object,
    ) -> _Linker:
        captured["loader"] = contract_loader
        return _Linker()

    monkeypatch.setattr(
        "dc43_service_backends.governance.unity_catalog.build_linker_from_config",
        _fake_builder,
    )

    config = ServiceBackendsConfig(
        unity_catalog=UnityCatalogConfig(enabled=True, sql_dsn="sqlite:///tmp.db"),
    )
    context = LinkHookContext(contract_service=_ContractService())

    hooks = build_link_hooks(config, context=context)

    assert hooks is not None and len(hooks) == 1
    loader = captured.get("loader")
    assert callable(loader)
    loader("sales.orders", "0.1.0")
    assert captured["requested"] == ("sales.orders", "0.1.0")


def test_build_linker_from_config_disabled() -> None:
    config = UnityCatalogConfig(enabled=False)
    linker = build_linker_from_config(config)
    assert linker is None


def test_build_dataset_contract_link_hooks_uses_unity(monkeypatch) -> None:
    config = ServiceBackendsConfig()
    config.unity_catalog.enabled = True

    linker = UnityCatalogLinker(apply_table_properties=lambda name, props: None)

    monkeypatch.setattr(
        "dc43_service_backends.governance.unity_catalog.build_linker_from_config",
        lambda cfg, **kwargs: linker,
    )

    hooks = build_dataset_contract_link_hooks(config)
    assert hooks == (linker.link_dataset_contract,)


def test_build_dataset_contract_link_hooks_warns_on_failure() -> None:
    config = ServiceBackendsConfig()

    def _broken_builder(*_: object) -> None:
        raise RuntimeError("boom")

    with pytest.warns(RuntimeWarning):
        hooks = build_dataset_contract_link_hooks(
            config,
            extra_builders=[_broken_builder],
            include_defaults=False,
        )

    assert hooks == ()


def test_build_dataset_contract_link_hooks_pass_context() -> None:
    config = ServiceBackendsConfig()
    context = LinkHookContext(contract_service=SimpleNamespace())
    captured: dict[str, object] = {}

    def _builder(_: ServiceBackendsConfig, context: LinkHookContext | None = None):
        captured["context"] = context
        return ()

    hooks = build_dataset_contract_link_hooks(
        config,
        include_defaults=False,
        extra_builders=[_builder],
        context=context,
    )

    assert hooks == ()
    assert captured["context"] is context


def test_build_dataset_contract_link_hooks_uses_configured_specs(monkeypatch) -> None:
    config = ServiceBackendsConfig()
    config.governance.dataset_contract_link_builders = (
        "custom.module:builder",
        "custom.module:builder",
    )

    loaded: list[str] = []

    def _loader(spec: str):
        loaded.append(spec)
        return lambda cfg, context=None: None

    monkeypatch.setattr(
        "dc43_service_backends.governance.bootstrap.load_link_hook_builder",
        _loader,
    )

    hooks = build_dataset_contract_link_hooks(config, include_defaults=False)
    assert hooks == ()
    assert loaded == ["custom.module:builder"]


def test_build_dataset_contract_link_hooks_warns_on_load_failure(monkeypatch) -> None:
    config = ServiceBackendsConfig()
    config.governance.dataset_contract_link_builders = ("broken.module:builder",)

    def _loader(spec: str):
        raise RuntimeError("load boom")

    monkeypatch.setattr(
        "dc43_service_backends.governance.bootstrap.load_link_hook_builder",
        _loader,
    )

    with pytest.warns(RuntimeWarning):
        hooks = build_dataset_contract_link_hooks(config, include_defaults=False)

    assert hooks == ()


def test_load_link_hook_builder_colon_spec() -> None:
    module = SimpleNamespace(builder=lambda config: ())
    sys.modules["sample.builders"] = module
    try:
        builder = load_link_hook_builder("sample.builders:builder")
        assert builder is module.builder
    finally:
        sys.modules.pop("sample.builders", None)


def test_load_link_hook_builder_dotted_spec() -> None:
    module = SimpleNamespace(nested=SimpleNamespace(factory=lambda config: ()))
    sys.modules["sample.builders2"] = module
    try:
        builder = load_link_hook_builder("sample.builders2.nested.factory")
        assert builder is module.nested.factory
    finally:
        sys.modules.pop("sample.builders2", None)


def test_default_builder_specs_include_unity() -> None:
    assert (
        "dc43_service_backends.governance.unity_catalog:build_link_hooks"
        in DEFAULT_LINK_HOOK_BUILDER_SPECS
    )



"""Compatibility shim exposing the relocated in-memory governance store."""

from __future__ import annotations

from ..backend.stores import memory as _memory  # type: ignore
from ..backend.stores.memory import *  # type: ignore[F401,F403]

try:
    __all__ = list(_memory.__all__)
except AttributeError:
    __all__ = []


def __getattr__(name: str) -> object:
    """Delegate attribute access to the relocated module."""

    if name in _memory.__dict__:
        return _memory.__dict__[name]
    raise AttributeError(name)


def __dir__() -> list[str]:  # pragma: no cover - trivial passthrough
    """Combine the compatibility shim attributes with the target module ones."""

    return sorted(set(globals()) | set(dir(_memory)))



"""Protocol describing data product service operations."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Mapping, Optional, Protocol, Sequence

from dc43_service_clients.odps import (
    DataProductInputPort,
    DataProductOutputPort,
    OpenDataProductStandard,
)


@dataclass
class DataProductRegistrationResult:
    """Result returned by port registration operations."""

    product: OpenDataProductStandard
    changed: bool


@dataclass(slots=True)
class DataProductListing:
    """Paginated collection of data product identifiers."""

    items: Sequence[str]
    total: int
    limit: int | None = None
    offset: int = 0


class DataProductServiceBackend(Protocol):
    """Interface implemented by data product management backends."""

    def put(self, product: OpenDataProductStandard) -> None:
        """Persist ``product`` making it available for subsequent lookups."""

    def list_data_products(
        self, *, limit: int | None = None, offset: int = 0
    ) -> DataProductListing:
        """Return identifiers for data products tracked by the backend."""

    def get(self, data_product_id: str, version: str) -> OpenDataProductStandard:
        """Return a specific version of ``data_product_id``."""

    def latest(self, data_product_id: str) -> Optional[OpenDataProductStandard]:
        """Return the latest version of ``data_product_id`` when available."""

    def list_versions(self, data_product_id: str) -> Sequence[str]:
        """Return the known versions for ``data_product_id``."""

    def register_input_port(
        self,
        *,
        data_product_id: str,
        port: DataProductInputPort,
        bump: str = "minor",
        custom_properties: Optional[Mapping[str, object]] = None,
        source_data_product: Optional[str] = None,
        source_output_port: Optional[str] = None,
    ) -> DataProductRegistrationResult:
        """Ensure ``port`` is attached to ``data_product_id`` input ports."""

    def register_output_port(
        self,
        *,
        data_product_id: str,
        port: DataProductOutputPort,
        bump: str = "minor",
        custom_properties: Optional[Mapping[str, object]] = None,
    ) -> DataProductRegistrationResult:
        """Ensure ``port`` is attached to ``data_product_id`` output ports."""

    def resolve_output_contract(
        self,
        *,
        data_product_id: str,
        port_name: str,
    ) -> Optional[tuple[str, str]]:
        """Return ``(contract_id, contract_version)`` for ``port_name`` when known."""


__all__ = [
    "DataProductRegistrationResult",
    "DataProductListing",
    "DataProductServiceBackend",
]




"""Persistence adapters for data product definitions."""

from .interface import DataProductStore
from .memory import InMemoryDataProductStore
from .filesystem import FilesystemDataProductStore

try:  # pragma: no cover - optional dependency
    from .sql import SQLDataProductStore  # type: ignore
except ModuleNotFoundError:  # pragma: no cover - sqlalchemy optional
    SQLDataProductStore = None  # type: ignore[assignment]

try:  # pragma: no cover - optional dependency
    from .delta import DeltaDataProductStore  # type: ignore
except ModuleNotFoundError:  # pragma: no cover - pyspark optional
    DeltaDataProductStore = None  # type: ignore[assignment]

__all__ = [
    "DataProductStore",
    "InMemoryDataProductStore",
    "FilesystemDataProductStore",
    "SQLDataProductStore",
    "DeltaDataProductStore",
]



"""Helpers for constructing HTTP applications with local backends."""

from __future__ import annotations

try:  # pragma: no cover - optional dependency guard
    from fastapi import FastAPI
except ModuleNotFoundError as exc:  # pragma: no cover
    raise ModuleNotFoundError(
        "FastAPI is required to build HTTP applications. Install "
        "'dc43-service-backends[http]' to enable this helper."
    ) from exc

from typing import Sequence

from dc43_service_backends.contracts.backend import ContractServiceBackend, LocalContractServiceBackend
from dc43_service_backends.contracts.backend.stores import FSContractStore
from dc43_service_backends.contracts.backend.stores.interface import ContractStore
from dc43_service_backends.data_products import (
    DataProductServiceBackend,
    LocalDataProductServiceBackend,
)
from dc43_service_backends.data_quality.backend import (
    DataQualityServiceBackend,
    LocalDataQualityServiceBackend,
)
from dc43_service_backends.governance.backend import GovernanceServiceBackend, LocalGovernanceServiceBackend
from dc43_service_backends.governance.storage import GovernanceStore, InMemoryGovernanceStore
from dc43_service_backends.governance.hooks import DatasetContractLinkHook

from .server import build_app


def build_local_app(
    store: ContractStore | str,
    *,
    contract_backend: ContractServiceBackend | None = None,
    data_product_backend: DataProductServiceBackend | None = None,
    dq_backend: DataQualityServiceBackend | None = None,
    governance_backend: GovernanceServiceBackend | None = None,
    governance_store: GovernanceStore | None = None,
    link_hooks: Sequence[DatasetContractLinkHook] | None = None,
    dependencies: Sequence[object] | None = None,
) -> FastAPI:
    """Build a FastAPI application wired against in-process backends."""

    if isinstance(store, str):
        store = FSContractStore(store)

    if contract_backend is None:
        contract_backend = LocalContractServiceBackend(store)
    if data_product_backend is None:
        data_product_backend = LocalDataProductServiceBackend()
    if dq_backend is None:
        dq_backend = LocalDataQualityServiceBackend()
    if governance_store is None:
        governance_store = InMemoryGovernanceStore()
    if governance_backend is None:
        governance_backend = LocalGovernanceServiceBackend(
            contract_client=contract_backend,
            dq_client=dq_backend,
            draft_store=store,
            link_hooks=link_hooks,
            store=governance_store,
        )

    return build_app(
        contract_backend=contract_backend,
        data_product_backend=data_product_backend,
        dq_backend=dq_backend,
        governance_backend=governance_backend,
        dependencies=dependencies,
    )


__all__ = ["build_local_app"]



"""Service-side contract for governance orchestration."""

from __future__ import annotations

from typing import Callable, Mapping, Optional, Protocol, Sequence

from open_data_contract_standard.model import OpenDataContractStandard  # type: ignore

from dc43_service_clients.data_quality import ObservationPayload, ValidationResult
from dc43_service_clients.governance.lineage import OpenDataLineageEvent
from dc43_service_clients.governance.models import (
    GovernanceReadContext,
    GovernanceWriteContext,
    GovernanceCredentials,
    PipelineContextSpec,
    QualityAssessment,
    QualityDraftContext,
    ResolvedReadPlan,
    ResolvedWritePlan,
)


class GovernanceServiceBackend(Protocol):
    """Operations exposed by a governance service implementation."""

    def get_contract(
        self,
        *,
        contract_id: str,
        contract_version: str,
    ) -> OpenDataContractStandard:
        ...

    def latest_contract(
        self,
        *,
        contract_id: str,
    ) -> Optional[OpenDataContractStandard]:
        ...

    def list_contract_versions(self, *, contract_id: str) -> Sequence[str]:
        ...

    def describe_expectations(
        self,
        *,
        contract_id: str,
        contract_version: str,
    ) -> Sequence[Mapping[str, object]]:
        ...

    def configure_auth(
        self,
        credentials: GovernanceCredentials | Mapping[str, object] | str | None,
    ) -> None:
        ...

    def evaluate_dataset(
        self,
        *,
        contract_id: str,
        contract_version: str,
        dataset_id: str,
        dataset_version: str,
        validation: ValidationResult | None,
        observations: Callable[[], ObservationPayload],
        bump: str = "minor",
        context: QualityDraftContext | None = None,
        pipeline_context: PipelineContextSpec | None = None,
        operation: str = "read",
        draft_on_violation: bool = False,
    ) -> QualityAssessment:
        ...

    def review_validation_outcome(
        self,
        *,
        validation: ValidationResult,
        base_contract: OpenDataContractStandard,
        bump: str = "minor",
        dataset_id: str | None = None,
        dataset_version: str | None = None,
        data_format: str | None = None,
        dq_status: ValidationResult | None = None,
        dq_feedback: Mapping[str, object] | None = None,
        context: QualityDraftContext | None = None,
        pipeline_context: PipelineContextSpec | None = None,
        draft_requested: bool = False,
        operation: str | None = None,
    ) -> Optional[OpenDataContractStandard]:
        ...

    def propose_draft(
        self,
        *,
        validation: ValidationResult,
        base_contract: OpenDataContractStandard,
        bump: str = "minor",
        context: QualityDraftContext | None = None,
        pipeline_context: PipelineContextSpec | None = None,
    ) -> OpenDataContractStandard:
        ...

    def get_status(
        self,
        *,
        contract_id: str,
        contract_version: str,
        dataset_id: str,
        dataset_version: str,
    ) -> Optional[ValidationResult]:
        ...

    def link_dataset_contract(
        self,
        *,
        dataset_id: str,
        dataset_version: str,
        contract_id: str,
        contract_version: str,
    ) -> None:
        ...

    def get_linked_contract_version(
        self,
        *,
        dataset_id: str,
        dataset_version: Optional[str] = None,
    ) -> Optional[str]:
        ...

    def get_metrics(
        self,
        *,
        dataset_id: str,
        dataset_version: Optional[str] = None,
        contract_id: Optional[str] = None,
        contract_version: Optional[str] = None,
    ) -> Sequence[Mapping[str, object]]:
        ...

    def get_status_matrix(
        self,
        *,
        dataset_id: str,
        contract_ids: Sequence[str] | None = None,
        dataset_versions: Sequence[str] | None = None,
    ) -> Sequence[Mapping[str, object]]:
        ...

    def list_datasets(self) -> Sequence[str]:
        ...

    def get_pipeline_activity(
        self,
        *,
        dataset_id: str,
        dataset_version: Optional[str] = None,
        include_status: bool = False,
    ) -> Sequence[Mapping[str, object]]:
        ...

    def get_dataset_records(
        self,
        *,
        dataset_id: str | None = None,
        dataset_version: str | None = None,
    ) -> Sequence[Mapping[str, object]]:
        ...

    def resolve_read_context(
        self,
        *,
        context: GovernanceReadContext,
    ) -> ResolvedReadPlan:
        ...

    def resolve_write_context(
        self,
        *,
        context: GovernanceWriteContext,
    ) -> ResolvedWritePlan:
        ...

    def evaluate_read_plan(
        self,
        *,
        plan: ResolvedReadPlan,
        validation: ValidationResult | None,
        observations: Callable[[], ObservationPayload],
    ) -> QualityAssessment:
        ...

    def evaluate_write_plan(
        self,
        *,
        plan: ResolvedWritePlan,
        validation: ValidationResult | None,
        observations: Callable[[], ObservationPayload],
    ) -> QualityAssessment:
        ...

    def register_read_activity(
        self,
        *,
        plan: ResolvedReadPlan,
        assessment: QualityAssessment,
    ) -> None:
        ...

    def register_write_activity(
        self,
        *,
        plan: ResolvedWritePlan,
        assessment: QualityAssessment,
    ) -> None:
        ...

    def publish_lineage_event(
        self,
        *,
        event: OpenDataLineageEvent,
    ) -> None:
        ...


__all__ = ["GovernanceServiceBackend"]



"""Delta Lake backed persistence for ODPS documents."""

from __future__ import annotations

import logging
from typing import Iterable, Optional, Sequence

try:  # pragma: no cover - optional dependency
    from pyspark.sql import SparkSession
except Exception:  # pragma: no cover - Spark is optional for most deployments
    SparkSession = object  # type: ignore

from dc43_service_clients.odps import OpenDataProductStandard, as_odps_dict, to_model

from .._shared import _version_key
from .interface import DataProductStore


logger = logging.getLogger(__name__)


class DeltaDataProductStore(DataProductStore):
    """Persist ODPS documents in a Delta table or Unity Catalog object."""

    def __init__(
        self,
        spark: SparkSession,
        *,
        table: str | None = None,
        path: str | None = None,
        log_sql: bool = False,
    ) -> None:
        if not (table or path):
            raise ValueError("Provide either a Unity Catalog table name or a Delta path")
        self._spark = spark
        self._table = table
        self._path = path
        self._log_sql = log_sql
        self._ensure_table()

    # ------------------------------------------------------------------
    # Helpers
    # ------------------------------------------------------------------
    def _table_ref(self) -> str:
        return self._table if self._table else f"delta.`{self._path}`"

    def _execute_sql(self, statement: str):  # pragma: no cover - logging shim
        if self._log_sql:
            logger.info("Spark SQL (data products): %s", statement.strip())
        return self._spark.sql(statement)

    def _ensure_table(self) -> None:
        ref = self._table_ref()
        if self._table:
            self._execute_sql(
                f"""
                CREATE TABLE IF NOT EXISTS {ref} (
                    data_product_id STRING,
                    version STRING,
                    status STRING,
                    json STRING,
                    updated_at TIMESTAMP
                ) USING DELTA
                PARTITIONED BY (data_product_id)
                TBLPROPERTIES (delta.autoOptimize.autoCompact = true)
                """
            )
        else:
            self._execute_sql(
                f"""
                CREATE TABLE IF NOT EXISTS {ref} (
                    data_product_id STRING,
                    version STRING,
                    status STRING,
                    json STRING,
                    updated_at TIMESTAMP
                ) USING DELTA
                LOCATION '{self._path}'
                PARTITIONED BY (data_product_id)
                TBLPROPERTIES (delta.autoOptimize.autoCompact = true)
                """
            )

    def _merge_row(self, product: OpenDataProductStandard) -> None:
        import json

        ref = self._table_ref()
        payload = as_odps_dict(product)
        json_payload = json.dumps(payload, separators=(",", ":"))
        status = product.status or "draft"
        json_sql = json_payload.replace("'", "''")
        status_sql = status.replace("'", "''")
        self._execute_sql(
            f"""
            MERGE INTO {ref} t
            USING (SELECT
                    '{product.id}' as data_product_id,
                    '{product.version}' as version,
                    '{status_sql}' as status,
                    '{json_sql}' as json,
                    current_timestamp() as updated_at) s
            ON t.data_product_id = s.data_product_id AND t.version = s.version
            WHEN MATCHED THEN UPDATE SET status = s.status, json = s.json, updated_at = s.updated_at
            WHEN NOT MATCHED THEN INSERT *
            """
        )

    @staticmethod
    def _safe_json(payload: str) -> dict[str, object]:
        import json

        return json.loads(payload)

    def _collect_versions(self, data_product_id: str) -> Iterable[tuple[str, object]]:
        ref = self._table_ref()
        rows = self._execute_sql(
            f"SELECT version, json FROM {ref} WHERE data_product_id = '{data_product_id}'"
        ).collect()
        for row in rows:
            if not row:
                continue
            version = str(row[0]).strip()
            if not version:
                continue
            yield version, row[1]

    # ------------------------------------------------------------------
    # DataProductStore implementation
    # ------------------------------------------------------------------
    def put(self, product: OpenDataProductStandard) -> None:  # noqa: D401 - short docstring
        if not product.version:
            raise ValueError("Data product version is required")
        self._merge_row(product)

    def get(self, data_product_id: str, version: str) -> OpenDataProductStandard:  # noqa: D401
        ref = self._table_ref()
        rows = self._execute_sql(
            f"SELECT json FROM {ref} WHERE data_product_id = '{data_product_id}' AND version = '{version}'"
        ).head(1)
        if not rows:
            raise FileNotFoundError(f"data product {data_product_id}:{version} not found")
        return to_model(self._safe_json(rows[0][0]))

    def latest(self, data_product_id: str) -> Optional[OpenDataProductStandard]:  # noqa: D401
        entries = list(self._collect_versions(data_product_id))
        if not entries:
            return None
        latest_row = max(entries, key=lambda row: _version_key(row[0]))
        return to_model(self._safe_json(latest_row[1]))

    def list_versions(self, data_product_id: str) -> Sequence[str]:  # noqa: D401
        versions = sorted(
            (version for version, _ in self._collect_versions(data_product_id)),
            key=_version_key,
        )
        return [version for version in versions if version]

    def list_data_product_ids(self) -> Sequence[str]:  # noqa: D401
        ref = self._table_ref()
        rows = self._execute_sql(f"SELECT DISTINCT data_product_id FROM {ref}").collect()
        product_ids = {str(row[0]) for row in rows if row and row[0]}
        return sorted(product_ids)


__all__ = ["DeltaDataProductStore"]



from __future__ import annotations

import json
from pathlib import Path

import pytest

pytest.importorskip("sqlalchemy")

from open_data_contract_standard.model import Description, OpenDataContractStandard, SchemaObject, SchemaProperty  # type: ignore

from dc43_service_backends.contracts.backend.stores.sql import SQLContractStore


def _build_contract(version: str = "1.0.0", description: str = "Orders facts") -> OpenDataContractStandard:
    return OpenDataContractStandard(
        version=version,
        kind="DataContract",
        apiVersion="3.0.2",
        id="test.orders",
        name="Orders",
        description=Description(usage=description),
        schema=[
            SchemaObject(
                name="orders",
                properties=[
                    SchemaProperty(name="order_id", physicalType="bigint", required=True),
                ],
            )
        ],
    )


@pytest.fixture()
def sql_engine(tmp_path: Path):
    pytest.importorskip("sqlalchemy")
    from sqlalchemy import create_engine

    db_path = tmp_path / "contracts.db"
    engine = create_engine(f"sqlite:///{db_path}")
    yield engine
    engine.dispose()


def test_sql_contract_store_round_trip(sql_engine) -> None:
    store = SQLContractStore(sql_engine)
    contract = _build_contract()

    store.put(contract)

    retrieved = store.get("test.orders", "1.0.0")
    assert retrieved.id == "test.orders"
    assert retrieved.description
    assert retrieved.description.usage == "Orders facts"

    contracts = store.list_contracts()
    assert contracts == ["test.orders"]

    versions = store.list_versions("test.orders")
    assert versions == ["1.0.0"]


def test_sql_contract_store_updates_existing_version(sql_engine) -> None:
    store = SQLContractStore(sql_engine)
    original = _build_contract()
    updated = _build_contract(description="Orders facts v2")

    store.put(original)
    store.put(updated)

    latest = store.get("test.orders", "1.0.0")
    assert latest.description
    assert latest.description.usage == "Orders facts v2"


def test_sql_contract_store_serialises_contract(sql_engine) -> None:
    store = SQLContractStore(sql_engine)
    contract = _build_contract()

    store.put(contract)

    from sqlalchemy import text

    with sql_engine.begin() as conn:
        payload = conn.execute(
            text("SELECT json FROM contracts WHERE contract_id = :cid AND version = :version"),
            {"cid": "test.orders", "version": "1.0.0"},
        ).scalar_one()

    decoded = json.loads(payload)
    assert decoded["id"] == "test.orders"
    assert decoded["version"] == "1.0.0"



"""HTTP application exposing service backend capabilities."""

from __future__ import annotations

from typing import Any, List, Mapping, Optional, Sequence

try:  # pragma: no cover - import guard exercised in packaging contexts
    from fastapi import APIRouter, FastAPI, HTTPException, Query, Response
    from fastapi.responses import RedirectResponse
    from fastapi.encoders import jsonable_encoder
except ModuleNotFoundError as exc:  # pragma: no cover - raised when optional deps missing
    raise ModuleNotFoundError(
        "FastAPI is required to use the HTTP server utilities. Install "
        "'dc43-service-backends[http]' to enable them."
    ) from exc
from pydantic import BaseModel
from open_data_contract_standard.model import OpenDataContractStandard  # type: ignore

from dc43_service_clients.odps import (
    DataProductInputPort,
    DataProductOutputPort,
    as_odps_dict as as_odps_product_dict,
    to_model as to_data_product_model,
)
from dc43_service_clients.data_quality import ValidationResult
from dc43_service_clients.data_quality.transport import (
    decode_observation_payload,
    decode_validation_result,
    encode_validation_result,
)
from dc43_service_clients.governance.lineage import decode_lineage_event
from dc43_service_clients.governance.transport import (
    decode_contract,
    decode_credentials,
    decode_draft_context,
    decode_quality_assessment,
    decode_read_context,
    decode_read_plan,
    decode_write_context,
    decode_write_plan,
    encode_contract,
    encode_quality_assessment,
    encode_read_plan,
    encode_write_plan,
)
from dc43_service_clients.governance.models import QualityAssessment

from .contracts import ContractServiceBackend
from .data_products import DataProductServiceBackend
from .data_quality import DataQualityServiceBackend
from .governance.backend import GovernanceServiceBackend
from .core.odcs import to_model as to_contract_model


class _LinkDatasetPayload(BaseModel):
    dataset_id: str
    dataset_version: str
    contract_id: str
    contract_version: str


class _DataProductInputPayload(BaseModel):
    port_name: str
    contract_id: str
    contract_version: str
    bump: str = "minor"
    custom_properties: Optional[Mapping[str, Any]] = None
    source_data_product: Optional[str] = None
    source_output_port: Optional[str] = None


class _DataProductOutputPayload(BaseModel):
    port_name: str
    contract_id: str
    contract_version: str
    bump: str = "minor"
    custom_properties: Optional[Mapping[str, Any]] = None


class _EvaluateDQPayload(BaseModel):
    contract: Mapping[str, Any]
    payload: Mapping[str, Any]


class _ExpectationsPayload(BaseModel):
    contract: Mapping[str, Any]


class _GovernanceEvaluatePayload(BaseModel):
    contract_id: str
    contract_version: str
    dataset_id: str
    dataset_version: str
    validation: Optional[Mapping[str, Any]] = None
    observations: Mapping[str, Any]
    bump: str = "minor"
    context: Optional[Mapping[str, Any]] = None
    pipeline_context: Optional[Mapping[str, Any]] = None
    operation: str = "read"
    draft_on_violation: bool = False


class _GovernanceReviewPayload(BaseModel):
    validation: Mapping[str, Any]
    base_contract: Mapping[str, Any]
    bump: str = "minor"
    dataset_id: Optional[str] = None
    dataset_version: Optional[str] = None
    data_format: Optional[str] = None
    dq_status: Optional[Mapping[str, Any]] = None
    dq_feedback: Optional[Mapping[str, Any]] = None
    context: Optional[Mapping[str, Any]] = None
    pipeline_context: Optional[Mapping[str, Any]] = None
    draft_requested: bool = False
    operation: Optional[str] = None


class _GovernanceDraftPayload(BaseModel):
    validation: Mapping[str, Any]
    base_contract: Mapping[str, Any]
    bump: str = "minor"
    context: Optional[Mapping[str, Any]] = None
    pipeline_context: Optional[Mapping[str, Any]] = None


class _AuthPayload(BaseModel):
    credentials: Optional[Mapping[str, Any]] = None


class _GovernanceResolvePayload(BaseModel):
    context: Mapping[str, Any]


class _GovernancePlanEvaluatePayload(BaseModel):
    plan: Mapping[str, Any]
    validation: Optional[Mapping[str, Any]] = None
    observations: Optional[Mapping[str, Any]] = None


class _GovernanceRegisterPayload(BaseModel):
    plan: Mapping[str, Any]
    assessment: Mapping[str, Any]


class _GovernanceLineagePayload(BaseModel):
    event: Mapping[str, Any]


def _encode_assessment(assessment: QualityAssessment) -> Mapping[str, Any]:
    return encode_quality_assessment(assessment)


def build_app(
    *,
    contract_backend: ContractServiceBackend,
    dq_backend: DataQualityServiceBackend,
    governance_backend: GovernanceServiceBackend,
    data_product_backend: DataProductServiceBackend,
    dependencies: Sequence[object] | None = None,
) -> FastAPI:
    """Create a FastAPI app exposing the provided backend implementations."""

    app = FastAPI(title="dc43 service backends")
    router_dependencies = list(dependencies) if dependencies else None
    router = APIRouter(dependencies=router_dependencies)

    @app.get("/", include_in_schema=False)
    def docs_redirect() -> Response:
        """Expose the interactive API documentation at the application root."""

        if app.docs_url:
            return RedirectResponse(url=app.docs_url)
        if app.openapi_url:
            return RedirectResponse(url=app.openapi_url)
        return Response(status_code=204)

    # ------------------------------------------------------------------
    # Contract service endpoints
    # ------------------------------------------------------------------
    @router.put("/contracts/{contract_id}/versions/{contract_version}", status_code=204)
    def put_contract(contract_id: str, contract_version: str, payload: Mapping[str, Any]) -> None:
        try:
            contract = to_contract_model(dict(payload))
        except Exception as exc:  # pragma: no cover - invalid payload handling
            raise HTTPException(status_code=400, detail=f"invalid contract payload: {exc}") from exc

        payload_id = str(contract.id) if contract.id else None
        if payload_id and payload_id != contract_id:
            raise HTTPException(status_code=400, detail="contract.id does not match request path")

        payload_version = str(contract.version) if contract.version else None
        if payload_version and payload_version != contract_version:
            raise HTTPException(status_code=400, detail="contract.version does not match request path")

        contract.id = contract_id
        contract.version = contract_version
        contract_backend.put(contract)

    @router.get("/contracts")
    def list_contracts(
        limit: int | None = Query(None, ge=0),
        offset: int = Query(0, ge=0),
    ) -> Mapping[str, Any]:
        listing = contract_backend.list_contracts(limit=limit, offset=offset)
        return {
            "items": [str(value) for value in listing.items],
            "total": int(listing.total),
            "limit": listing.limit,
            "offset": listing.offset,
        }

    @router.get("/contracts/{contract_id}/versions/{contract_version}")
    def get_contract(contract_id: str, contract_version: str) -> Mapping[str, Any]:
        try:
            contract = contract_backend.get(contract_id, contract_version)
        except FileNotFoundError as exc:  # pragma: no cover - backend signals missing contract
            raise HTTPException(status_code=404, detail=str(exc)) from exc
        return contract.model_dump(by_alias=True, exclude_none=True)

    @router.get("/contracts/{contract_id}/latest")
    def latest_contract(contract_id: str) -> Mapping[str, Any]:
        contract = contract_backend.latest(contract_id)
        if contract is None:
            raise HTTPException(status_code=404, detail="contract not found")
        return contract.model_dump(by_alias=True, exclude_none=True)

    @router.get("/contracts/{contract_id}/versions")
    def list_contract_versions(contract_id: str) -> list[str]:
        versions = contract_backend.list_versions(contract_id)
        return [str(value) for value in versions]

    @router.post("/contracts/link")
    def link_contract(payload: _LinkDatasetPayload) -> None:
        contract_backend.link_dataset_contract(
            dataset_id=payload.dataset_id,
            dataset_version=payload.dataset_version,
            contract_id=payload.contract_id,
            contract_version=payload.contract_version,
        )

    @router.get("/contracts/datasets/{dataset_id}/linked")
    def get_linked_contract(dataset_id: str, dataset_version: str | None = None) -> Mapping[str, Any]:
        version = contract_backend.get_linked_contract_version(
            dataset_id=dataset_id,
            dataset_version=dataset_version,
        )
        if version is None:
            raise HTTPException(status_code=404, detail="no contract linked")
        return {"contract_version": version}

    # ------------------------------------------------------------------
    # Data product endpoints
    # ------------------------------------------------------------------
    @router.put("/data-products/{data_product_id}/versions/{version}", status_code=204)
    def put_data_product(data_product_id: str, version: str, payload: Mapping[str, Any]) -> None:
        try:
            product = to_data_product_model(dict(payload))
        except Exception as exc:  # pragma: no cover - invalid payload handling
            raise HTTPException(status_code=400, detail=f"invalid data product payload: {exc}") from exc

        payload_id = str(product.id) if product.id else None
        if payload_id and payload_id != data_product_id:
            raise HTTPException(status_code=400, detail="data_product.id does not match request path")

        payload_version = str(product.version) if product.version else None
        if payload_version and payload_version != version:
            raise HTTPException(status_code=400, detail="data_product.version does not match request path")

        product.id = data_product_id
        product.version = version
        data_product_backend.put(product)

    @router.get("/data-products")
    def list_data_products(
        limit: int | None = Query(None, ge=0),
        offset: int = Query(0, ge=0),
    ) -> Mapping[str, Any]:
        try:
            listing = data_product_backend.list_data_products(limit=limit, offset=offset)
        except NotImplementedError as exc:  # pragma: no cover - backend does not support listings
            raise HTTPException(status_code=501, detail=str(exc)) from exc
        return {
            "items": [str(value) for value in listing.items],
            "total": int(listing.total),
            "limit": listing.limit,
            "offset": listing.offset,
        }

    @router.get("/data-products/{data_product_id}/versions/{version}")
    def get_data_product(data_product_id: str, version: str) -> Mapping[str, Any]:
        try:
            product = data_product_backend.get(data_product_id, version)
        except FileNotFoundError as exc:  # pragma: no cover - backend signals missing product
            raise HTTPException(status_code=404, detail=str(exc)) from exc
        return as_odps_product_dict(product)

    @router.get("/data-products/{data_product_id}/latest")
    def latest_data_product(data_product_id: str) -> Mapping[str, Any]:
        product = data_product_backend.latest(data_product_id)
        if product is None:
            raise HTTPException(status_code=404, detail="data product not found")
        return as_odps_product_dict(product)

    @router.get("/data-products/{data_product_id}/versions")
    def list_data_product_versions(data_product_id: str) -> list[str]:
        versions = data_product_backend.list_versions(data_product_id)
        return [str(value) for value in versions]

    @router.post("/data-products/{data_product_id}/input-ports")
    def register_data_product_input(
        data_product_id: str, payload: _DataProductInputPayload
    ) -> Mapping[str, Any]:
        result = data_product_backend.register_input_port(
            data_product_id=data_product_id,
            port=DataProductInputPort(
                name=payload.port_name,
                version=payload.contract_version,
                contract_id=payload.contract_id,
            ),
            bump=payload.bump,
            custom_properties=payload.custom_properties,
            source_data_product=payload.source_data_product,
            source_output_port=payload.source_output_port,
        )
        return {
            "product": as_odps_product_dict(result.product),
            "changed": result.changed,
        }

    @router.post("/data-products/{data_product_id}/output-ports")
    def register_data_product_output(
        data_product_id: str, payload: _DataProductOutputPayload
    ) -> Mapping[str, Any]:
        result = data_product_backend.register_output_port(
            data_product_id=data_product_id,
            port=DataProductOutputPort(
                name=payload.port_name,
                version=payload.contract_version,
                contract_id=payload.contract_id,
            ),
            bump=payload.bump,
            custom_properties=payload.custom_properties,
        )
        return {
            "product": as_odps_product_dict(result.product),
            "changed": result.changed,
        }

    @router.get("/data-products/{data_product_id}/output-ports/{port_name}/contract")
    def resolve_data_product_output_contract(
        data_product_id: str, port_name: str
    ) -> Mapping[str, Any]:
        contract = data_product_backend.resolve_output_contract(
            data_product_id=data_product_id,
            port_name=port_name,
        )
        if contract is None:
            raise HTTPException(status_code=404, detail="output port not found")
        contract_id, contract_version = contract
        return {
            "contract_id": contract_id,
            "contract_version": contract_version,
        }

    # ------------------------------------------------------------------
    # Data-quality endpoints
    # ------------------------------------------------------------------
    @router.post("/data-quality/evaluate")
    def evaluate_quality(payload: _EvaluateDQPayload) -> Mapping[str, Any]:
        contract = OpenDataContractStandard.model_validate(dict(payload.contract))
        observations = decode_observation_payload(payload.payload)
        result = dq_backend.evaluate(contract=contract, payload=observations)
        return encode_validation_result(result) or {}

    @router.post("/data-quality/expectations")
    def describe_expectations(payload: _ExpectationsPayload) -> list[Mapping[str, Any]]:
        contract = OpenDataContractStandard.model_validate(dict(payload.contract))
        descriptors = dq_backend.describe_expectations(contract=contract)
        return [dict(item) for item in descriptors]

    # ------------------------------------------------------------------
    # Governance endpoints
    # ------------------------------------------------------------------
    @router.post("/governance/auth")
    def configure_auth(payload: _AuthPayload) -> None:
        credentials = decode_credentials(payload.credentials)
        governance_backend.configure_auth(credentials)

    @router.post("/governance/read/resolve")
    def resolve_read_context(payload: _GovernanceResolvePayload) -> Mapping[str, Any]:
        context = decode_read_context(payload.context)
        if context is None:
            raise HTTPException(status_code=400, detail="invalid read context")
        plan = governance_backend.resolve_read_context(context=context)
        return encode_read_plan(plan)

    @router.post("/governance/write/resolve")
    def resolve_write_context(payload: _GovernanceResolvePayload) -> Mapping[str, Any]:
        context = decode_write_context(payload.context)
        if context is None:
            raise HTTPException(status_code=400, detail="invalid write context")
        plan = governance_backend.resolve_write_context(context=context)
        return encode_write_plan(plan)

    @router.post("/governance/read/evaluate")
    def evaluate_read_plan(payload: _GovernancePlanEvaluatePayload) -> Mapping[str, Any]:
        try:
            plan = decode_read_plan(payload.plan)
        except ValueError as exc:  # pragma: no cover - invalid payload handling
            raise HTTPException(status_code=400, detail=str(exc)) from exc
        validation = decode_validation_result(payload.validation)
        observations = decode_observation_payload(payload.observations or {})
        assessment = governance_backend.evaluate_read_plan(
            plan=plan,
            validation=validation,
            observations=lambda: observations,
        )
        return _encode_assessment(assessment)

    @router.post("/governance/write/evaluate")
    def evaluate_write_plan(payload: _GovernancePlanEvaluatePayload) -> Mapping[str, Any]:
        try:
            plan = decode_write_plan(payload.plan)
        except ValueError as exc:  # pragma: no cover - invalid payload handling
            raise HTTPException(status_code=400, detail=str(exc)) from exc
        validation = decode_validation_result(payload.validation)
        observations = decode_observation_payload(payload.observations or {})
        assessment = governance_backend.evaluate_write_plan(
            plan=plan,
            validation=validation,
            observations=lambda: observations,
        )
        return _encode_assessment(assessment)

    @router.post("/governance/read/register", status_code=204)
    def register_read_activity(payload: _GovernanceRegisterPayload) -> None:
        try:
            plan = decode_read_plan(payload.plan)
        except ValueError as exc:  # pragma: no cover - invalid payload handling
            raise HTTPException(status_code=400, detail=str(exc)) from exc
        assessment = decode_quality_assessment(payload.assessment)
        governance_backend.register_read_activity(plan=plan, assessment=assessment)

    @router.post("/governance/write/register", status_code=204)
    def register_write_activity(payload: _GovernanceRegisterPayload) -> None:
        try:
            plan = decode_write_plan(payload.plan)
        except ValueError as exc:  # pragma: no cover - invalid payload handling
            raise HTTPException(status_code=400, detail=str(exc)) from exc
        assessment = decode_quality_assessment(payload.assessment)
        governance_backend.register_write_activity(plan=plan, assessment=assessment)

    @router.post("/governance/lineage", status_code=204)
    def publish_lineage(payload: _GovernanceLineagePayload) -> None:
        try:
            event = decode_lineage_event(payload.event)
        except ValueError as exc:  # pragma: no cover - invalid payload handling
            raise HTTPException(status_code=400, detail=str(exc)) from exc
        if event is None:
            raise HTTPException(status_code=400, detail="lineage event payload is required")
        governance_backend.publish_lineage_event(event=event)

    @router.post("/governance/evaluate")
    def evaluate_dataset(payload: _GovernanceEvaluatePayload) -> Mapping[str, Any]:
        validation = decode_validation_result(payload.validation)
        observations = decode_observation_payload(payload.observations)
        context = decode_draft_context(payload.context)
        assessment = governance_backend.evaluate_dataset(
            contract_id=payload.contract_id,
            contract_version=payload.contract_version,
            dataset_id=payload.dataset_id,
            dataset_version=payload.dataset_version,
            validation=validation,
            observations=lambda: observations,
            bump=payload.bump,
            context=context,
            pipeline_context=payload.pipeline_context,
            operation=payload.operation,
            draft_on_violation=payload.draft_on_violation,
        )
        return _encode_assessment(assessment)

    @router.post("/governance/review")
    def review_validation(payload: _GovernanceReviewPayload) -> Mapping[str, Any] | None:
        validation = decode_validation_result(payload.validation)
        base_contract = OpenDataContractStandard.model_validate(dict(payload.base_contract))
        dq_status = decode_validation_result(payload.dq_status)
        context = decode_draft_context(payload.context)
        draft = governance_backend.review_validation_outcome(
            validation=validation,
            base_contract=base_contract,
            bump=payload.bump,
            dataset_id=payload.dataset_id,
            dataset_version=payload.dataset_version,
            data_format=payload.data_format,
            dq_status=dq_status,
            dq_feedback=payload.dq_feedback,
            context=context,
            pipeline_context=payload.pipeline_context,
            draft_requested=payload.draft_requested,
            operation=payload.operation,
        )
        return encode_contract(draft)

    @router.post("/governance/draft")
    def propose_draft(payload: _GovernanceDraftPayload) -> Mapping[str, Any]:
        validation = decode_validation_result(payload.validation)
        base_contract = OpenDataContractStandard.model_validate(dict(payload.base_contract))
        context = decode_draft_context(payload.context)
        draft = governance_backend.propose_draft(
            validation=validation,
            base_contract=base_contract,
            bump=payload.bump,
            context=context,
            pipeline_context=payload.pipeline_context,
        )
        return encode_contract(draft) or {}

    @router.get("/governance/status")
    def get_status(
        contract_id: str,
        contract_version: str,
        dataset_id: str,
        dataset_version: str,
    ) -> Mapping[str, Any]:
        status = governance_backend.get_status(
            contract_id=contract_id,
            contract_version=contract_version,
            dataset_id=dataset_id,
            dataset_version=dataset_version,
        )
        if status is None:
            raise HTTPException(status_code=404, detail="status unavailable")
        return encode_validation_result(status) or {}

    @router.get("/governance/status-matrix")
    def get_status_matrix(
        dataset_id: str,
        contract_id: List[str] | None = None,
        dataset_version: List[str] | None = None,
    ) -> Mapping[str, Any]:
        contract_ids: Sequence[str] | None
        if contract_id is None:
            contract_ids = None
        elif isinstance(contract_id, (list, tuple, set)):
            contract_ids = [str(item) for item in contract_id if str(item)]
        else:
            value = str(contract_id)
            contract_ids = [value] if value else None

        dataset_versions: Sequence[str] | None
        if dataset_version is None:
            dataset_versions = None
        elif isinstance(dataset_version, (list, tuple, set)):
            dataset_versions = [str(item) for item in dataset_version if str(item)]
        else:
            value = str(dataset_version)
            dataset_versions = [value] if value else None

        records = governance_backend.get_status_matrix(
            dataset_id=dataset_id,
            contract_ids=contract_ids,
            dataset_versions=dataset_versions,
        )
        entries: List[Mapping[str, Any]] = []
        for record in records:
            if not isinstance(record, Mapping):
                continue
            status_payload = record.get("status")
            encoded = None
            if isinstance(status_payload, ValidationResult):
                encoded = encode_validation_result(status_payload)
            elif isinstance(status_payload, Mapping):
                encoded = dict(status_payload)
            elif status_payload is not None:
                try:
                    encoded = encode_validation_result(status_payload)  # type: ignore[arg-type]
                except Exception:  # pragma: no cover - defensive guard for unexpected payloads
                    encoded = None
            entries.append(
                {
                    "dataset_id": record.get("dataset_id"),
                    "dataset_version": record.get("dataset_version"),
                    "contract_id": record.get("contract_id"),
                    "contract_version": record.get("contract_version"),
                    "status": encoded,
                }
            )
        return {"dataset_id": dataset_id, "entries": entries}

    @router.post("/governance/link", status_code=204)
    def link_dataset(payload: _LinkDatasetPayload) -> None:
        governance_backend.link_dataset_contract(
            dataset_id=payload.dataset_id,
            dataset_version=payload.dataset_version,
            contract_id=payload.contract_id,
            contract_version=payload.contract_version,
        )

    @router.get("/governance/linked")
    def get_link(dataset_id: str, dataset_version: str | None = None) -> Mapping[str, Any]:
        version = governance_backend.get_linked_contract_version(
            dataset_id=dataset_id,
            dataset_version=dataset_version,
        )
        if version is None:
            raise HTTPException(status_code=404, detail="no contract linked")
        return {"contract_version": version}

    @router.get("/governance/metrics")
    def dataset_metrics(
        dataset_id: str,
        dataset_version: str | None = None,
        contract_id: str | None = None,
        contract_version: str | None = None,
    ) -> list[Mapping[str, Any]]:
        records = governance_backend.get_metrics(
            dataset_id=dataset_id,
            dataset_version=dataset_version,
            contract_id=contract_id,
            contract_version=contract_version,
        )
        return list(jsonable_encoder(records))

    @router.get("/governance/datasets")
    def dataset_ids() -> list[str]:
        return list(governance_backend.list_datasets())

    @router.get("/governance/activity")
    def pipeline_activity(
        dataset_id: str,
        dataset_version: str | None = None,
        include_status: bool = False,
    ) -> list[Mapping[str, Any]]:
        records = governance_backend.get_pipeline_activity(
            dataset_id=dataset_id,
            dataset_version=dataset_version,
            include_status=include_status,
        )
        if not include_status:
            return list(jsonable_encoder(records))

        entries: list[Mapping[str, Any]] = []
        for record in records:
            if not isinstance(record, Mapping):
                entries.append(record)
                continue
            entry = dict(record)
            status_payload = entry.get("validation_status")
            encoded_status = None
            if isinstance(status_payload, ValidationResult):
                encoded_status = encode_validation_result(status_payload)
            elif isinstance(status_payload, Mapping):
                encoded_status = dict(status_payload)
            elif status_payload is not None:
                try:
                    encoded_status = encode_validation_result(status_payload)  # type: ignore[arg-type]
                except Exception:  # pragma: no cover - defensive guard
                    encoded_status = None
            if encoded_status is not None:
                entry["validation_status"] = encoded_status
            entries.append(entry)
        return list(jsonable_encoder(entries))

    @router.get("/governance/dataset-records")
    def dataset_records(
        dataset_id: str | None = None,
        dataset_version: str | None = None,
    ) -> list[Mapping[str, Any]]:
        records = governance_backend.get_dataset_records(
            dataset_id=dataset_id,
            dataset_version=dataset_version,
        )
        return [dict(record) for record in records]

    app.include_router(router)
    return app


__all__ = ["build_app"]



from __future__ import annotations

"""Abstract interface for contract storage backends."""

from abc import ABC, abstractmethod
from typing import List, Optional, Any, Dict

from open_data_contract_standard.model import OpenDataContractStandard  # type: ignore


class ContractStore(ABC):
    """Interface for reading/writing ODCS contracts in a registry."""
    @abstractmethod
    def put(self, contract: OpenDataContractStandard) -> None:
        """Store or update a contract (ODCS dict/object)."""
        ...

    @abstractmethod
    def get(self, contract_id: str, version: str) -> OpenDataContractStandard:
        """Fetch a contract by identity, returning an ODCS model."""
        ...

    @abstractmethod
    def list_contracts(self) -> List[str]:
        """List all contract identifiers available in the store."""
        ...

    @abstractmethod
    def list_versions(self, contract_id: str) -> List[str]:
        """List available versions for a given contract id."""
        ...

    def latest(self, contract_id: str) -> Optional[OpenDataContractStandard]:
        """Return the latest version for the contract id, if any."""
        versions = self.list_versions(contract_id)
        if not versions:
            return None
        versions.sort(key=lambda v: tuple(int(x) for x in v.split(".")[:3]))
        return self.get(contract_id, versions[-1])


__all__ = ["ContractStore"]



"""Re-export ODPS helpers from :mod:`dc43_core`."""

from __future__ import annotations

from dc43_core.odps import *  # noqa: F401,F403
from dc43_core.odps import __all__ as _CORE_ALL

__all__ = list(_CORE_ALL)



"""Interfaces for running data-quality service backends."""

from __future__ import annotations

from typing import Mapping, Protocol, Sequence

from open_data_contract_standard.model import OpenDataContractStandard  # type: ignore

from dc43_service_clients.data_quality import ObservationPayload, ValidationResult


class DataQualityServiceBackend(Protocol):
    """Service-side contract for evaluating data-quality observations."""

    def evaluate(
        self,
        *,
        contract: OpenDataContractStandard,
        payload: ObservationPayload,
    ) -> ValidationResult:
        """Return the validation outcome for the provided observations."""

    def describe_expectations(
        self, *, contract: OpenDataContractStandard
    ) -> Sequence[Mapping[str, object]]:
        """Return serialisable descriptors for contract expectations."""


__all__ = ["DataQualityServiceBackend"]



"""Persistence adapters for governance metadata and validation artefacts."""

from .interface import GovernanceStore
from .memory import InMemoryGovernanceStore
from .filesystem import FilesystemGovernanceStore

try:  # pragma: no cover - optional dependency
    from .sql import SQLGovernanceStore
except ModuleNotFoundError:  # pragma: no cover - sqlalchemy optional
    SQLGovernanceStore = None  # type: ignore[assignment]

try:  # pragma: no cover - optional dependencies
    from .delta import DeltaGovernanceStore
except ModuleNotFoundError:  # pragma: no cover - pyspark optional
    DeltaGovernanceStore = None  # type: ignore[assignment]

try:  # pragma: no cover - optional dependencies
    from .http import HttpGovernanceStore
except ModuleNotFoundError:  # pragma: no cover - httpx optional
    HttpGovernanceStore = None  # type: ignore[assignment]

__all__ = [
    "GovernanceStore",
    "InMemoryGovernanceStore",
    "FilesystemGovernanceStore",
    "SQLGovernanceStore",
    "DeltaGovernanceStore",
    "HttpGovernanceStore",
]



"""In-memory implementation of :class:`GovernanceStore`."""

from __future__ import annotations

import json
from collections import defaultdict
from datetime import datetime, timezone
from typing import Dict, List, Mapping, MutableMapping, Optional, Sequence

from dc43_service_clients.data_quality import ValidationResult

from ._metrics import extract_metrics, normalise_metric_value
from .interface import GovernanceStore


class InMemoryGovernanceStore(GovernanceStore):
    """Store governance artefacts in process memory."""

    def __init__(self) -> None:
        self._status_cache: Dict[tuple[str, str, str, str], ValidationResult] = {}
        self._dataset_links: Dict[tuple[str, Optional[str]], str] = {}
        self._activity_log: Dict[str, MutableMapping[str, MutableMapping[str, object]]] = (
            defaultdict(dict)
        )
        self._metrics: List[Dict[str, object]] = []

    # ------------------------------------------------------------------
    # Status persistence
    # ------------------------------------------------------------------
    def save_status(
        self,
        *,
        contract_id: str,
        contract_version: str,
        dataset_id: str,
        dataset_version: str,
        status: ValidationResult | None,
    ) -> None:
        key = (contract_id, contract_version, dataset_id, dataset_version)
        if status is None:
            self._status_cache.pop(key, None)
            return

        recorded_at = datetime.now(timezone.utc).isoformat().replace("+00:00", "Z")
        self._status_cache[key] = status
        metrics_map = extract_metrics(status)
        if metrics_map:
            self._record_metrics(
                contract_id=contract_id,
                contract_version=contract_version,
                dataset_id=dataset_id,
                dataset_version=dataset_version,
                recorded_at=recorded_at,
                metrics=metrics_map,
            )

    def load_status(
        self,
        *,
        contract_id: str,
        contract_version: str,
        dataset_id: str,
        dataset_version: str,
    ) -> ValidationResult | None:
        return self._status_cache.get((contract_id, contract_version, dataset_id, dataset_version))

    # ------------------------------------------------------------------
    # Dataset links
    # ------------------------------------------------------------------
    def link_dataset_contract(
        self,
        *,
        dataset_id: str,
        dataset_version: str,
        contract_id: str,
        contract_version: str,
    ) -> None:
        link_value = f"{contract_id}:{contract_version}"
        self._dataset_links[(dataset_id, dataset_version)] = link_value
        if dataset_version:
            self._dataset_links.setdefault((dataset_id, None), link_value)

    def get_linked_contract_version(
        self,
        *,
        dataset_id: str,
        dataset_version: Optional[str] = None,
    ) -> str | None:
        if (dataset_id, dataset_version) in self._dataset_links:
            return self._dataset_links[(dataset_id, dataset_version)]
        if dataset_version is not None and (dataset_id, None) in self._dataset_links:
            return self._dataset_links[(dataset_id, None)]
        for (linked_id, linked_version), value in self._dataset_links.items():
            if linked_id != dataset_id:
                continue
            if dataset_version is None or linked_version == dataset_version:
                return value
        return None

    # ------------------------------------------------------------------
    # Metrics
    # ------------------------------------------------------------------
    def _record_metrics(
        self,
        *,
        contract_id: str,
        contract_version: str,
        dataset_id: str,
        dataset_version: str,
        recorded_at: str,
        metrics: Mapping[str, object],
    ) -> None:
        for metric_key, metric_value in metrics.items():
            value, numeric = normalise_metric_value(metric_value)
            self._metrics.append(
                {
                    "dataset_id": dataset_id,
                    "dataset_version": dataset_version,
                    "contract_id": contract_id,
                    "contract_version": contract_version,
                    "status_recorded_at": recorded_at,
                    "metric_key": str(metric_key),
                    "metric_value": value,
                    "metric_numeric_value": numeric,
                }
            )

    def load_metrics(
        self,
        *,
        dataset_id: str,
        dataset_version: Optional[str] = None,
        contract_id: Optional[str] = None,
        contract_version: Optional[str] = None,
    ) -> Sequence[Mapping[str, object]]:
        records: list[Dict[str, object]] = []
        for entry in self._metrics:
            if entry.get("dataset_id") != dataset_id:
                continue
            if dataset_version is not None and entry.get("dataset_version") != dataset_version:
                continue
            if contract_id is not None and entry.get("contract_id") != contract_id:
                continue
            if contract_version is not None and entry.get("contract_version") != contract_version:
                continue
            records.append(dict(entry))

        records.sort(
            key=lambda item: (
                str(item.get("status_recorded_at", "")),
                str(item.get("metric_key", "")),
            )
        )
        return records

    # ------------------------------------------------------------------
    # Pipeline activity
    # ------------------------------------------------------------------
    def record_pipeline_event(
        self,
        *,
        contract_id: str,
        contract_version: str,
        dataset_id: str,
        dataset_version: str,
        event: Mapping[str, object],
        lineage_event: Mapping[str, object] | None = None,
    ) -> None:
        recorded_at = str(
            event.get("recorded_at")
            or datetime.now(timezone.utc).isoformat().replace("+00:00", "Z")
        )
        entry: Dict[str, object] = {
            "dataset_id": dataset_id,
            "dataset_version": dataset_version,
            "contract_id": contract_id,
            "contract_version": contract_version,
            "events": [],
        }
        record = self._activity_log[dataset_id].get(dataset_version)
        if isinstance(record, MutableMapping):
            entry.update(record)

        events = list(entry.get("events") or [])
        payload = dict(event)
        payload.setdefault("recorded_at", recorded_at)
        events.append(payload)
        if len(events) > 100:
            events = events[-100:]
        entry["events"] = events
        if lineage_event is not None:
            entry["lineage_event"] = dict(lineage_event)
        entry["contract_id"] = contract_id
        entry["contract_version"] = contract_version
        entry["dataset_id"] = dataset_id
        entry["dataset_version"] = dataset_version
        self._activity_log[dataset_id][dataset_version] = entry

    def list_datasets(self) -> Sequence[str]:
        dataset_ids = [dataset_id for dataset_id in self._activity_log.keys()]
        dataset_ids.sort()
        return dataset_ids

    def load_pipeline_activity(
        self,
        *,
        dataset_id: str,
        dataset_version: Optional[str] = None,
    ) -> Sequence[Mapping[str, object]]:
        dataset_entries = self._activity_log.get(dataset_id)
        if not dataset_entries:
            return []

        def _normalise(record: Mapping[str, object]) -> Dict[str, object]:
            entry = {
                "dataset_id": dataset_id,
                "dataset_version": record.get("dataset_version"),
                "contract_id": record.get("contract_id"),
                "contract_version": record.get("contract_version"),
                "events": [],
            }
            events = record.get("events")
            if isinstance(events, list):
                entry["events"] = [
                    dict(event)
                    for event in events
                    if isinstance(event, Mapping)
                ]
            lineage_payload = record.get("lineage_event")
            if isinstance(lineage_payload, Mapping):
                entry["lineage_event"] = dict(lineage_payload)
            return entry

        if dataset_version is not None:
            record = dataset_entries.get(dataset_version)
            if isinstance(record, Mapping):
                return [_normalise(record)]
            for candidate in dataset_entries.values():
                if (
                    isinstance(candidate, Mapping)
                    and candidate.get("dataset_version") == dataset_version
                ):
                    return [_normalise(candidate)]
            return []

        entries = [_normalise(record) for record in dataset_entries.values() if isinstance(record, Mapping)]

        def _sort_key(item: Mapping[str, object]) -> tuple[int, str]:
            events = item.get("events")
            recorded_at = ""
            if isinstance(events, list) and events:
                last = events[-1]
                if isinstance(last, Mapping):
                    recorded_at = str(last.get("recorded_at", ""))
            version_value = str(item.get("dataset_version", ""))
            return (0 if recorded_at else 1, recorded_at or version_value)

        entries.sort(key=_sort_key)
        return entries


__all__ = ["InMemoryGovernanceStore"]



"""Helpers to generate ODCS drafts from runtime observations."""

from __future__ import annotations

from datetime import datetime, timezone
import re
from typing import Any, Dict, Iterable, List, Mapping, Optional, Tuple
from uuid import uuid4

from open_data_contract_standard.model import (  # type: ignore
    CustomProperty,
    DataQuality,
    OpenDataContractStandard,
    SchemaProperty,
    Server,
)

from dc43_service_backends.core.odcs import as_odcs_dict, contract_identity, normalise_custom_properties, to_model
from dc43_service_clients.data_quality import ValidationResult
from dc43_service_backends.core.versioning import SemVer


_INVALID_IDENTIFIER = re.compile(r"[^0-9A-Za-z-]+")


def _normalise_identifier(value: str | None) -> Optional[str]:
    """Return a semver-friendly identifier derived from ``value``."""

    if value is None:
        return None
    token = _INVALID_IDENTIFIER.sub("-", str(value)).strip("-")
    return token or None


def _pipeline_hint(context: Mapping[str, Any] | None) -> Optional[str]:
    """Return a reviewer friendly label describing the draft origin."""

    if not context:
        return None

    for key in ("pipeline", "job", "project", "module", "function", "qualname", "source"):
        value = context.get(key)
        if value:
            token = _normalise_identifier(str(value))
            if token:
                return token
    return None


def _draft_version_suffix(
    *,
    dataset_id: Optional[str],
    dataset_version: Optional[str],
    draft_context: Optional[Mapping[str, Any]],
) -> str:
    """Return the pre-release suffix used to guarantee draft version uniqueness."""

    tokens: List[str] = ["draft"]

    for candidate in (dataset_version, dataset_id):
        token = _normalise_identifier(candidate)
        if token:
            tokens.append(token)

    pipeline_token = _pipeline_hint(draft_context)
    if pipeline_token:
        tokens.append(pipeline_token)

    timestamp = datetime.now(timezone.utc).strftime("%Y%m%dT%H%M%S%fZ")
    tokens.append(timestamp)

    entropy = uuid4().hex[:8]
    tokens.append(entropy)

    return "-".join(tokens)


def _resolve_observed_type(
    info: Mapping[str, Any] | None,
    fallback: str | None,
) -> Tuple[str, Optional[bool]]:
    """Return the preferred ODCS physical type and nullable flag."""

    observed_type = str(
        (info or {}).get("odcs_type")
        or (info or {}).get("type")
        or (info or {}).get("backend_type")
        or fallback
        or "string"
    )
    nullable = None
    if info is not None and "nullable" in info:
        nullable = bool(info.get("nullable", False))
    return observed_type, nullable


def _quality_rule_key(field: SchemaProperty, dq: DataQuality) -> Optional[Tuple[str, str]]:
    """Return the expectation rule prefix and human readable label."""

    name = field.name or ""
    if not name:
        return None

    if dq.mustBeGreaterThan is not None:
        return "gt", f"mustBeGreaterThan {dq.mustBeGreaterThan}"
    if dq.mustBeGreaterOrEqualTo is not None:
        return "ge", f"mustBeGreaterOrEqualTo {dq.mustBeGreaterOrEqualTo}"
    if dq.mustBeLessThan is not None:
        return "lt", f"mustBeLessThan {dq.mustBeLessThan}"
    if dq.mustBeLessOrEqualTo is not None:
        return "le", f"mustBeLessOrEqualTo {dq.mustBeLessOrEqualTo}"

    rule = (dq.rule or "").lower()
    if rule == "unique":
        return "unique", "unique"
    if rule == "enum" and isinstance(dq.mustBe, Iterable):
        return "enum", "enum"
    if rule == "regex" and dq.mustBe:
        return "regex", "regex"

    return None


def _quality_metric_value(
    *,
    metrics: Mapping[str, Any],
    rule_prefix: str,
    field_name: str,
) -> Optional[float]:
    key = f"violations.{rule_prefix}_{field_name}"
    value = metrics.get(key)
    if value is None:
        return None
    if isinstance(value, (int, float)):
        return float(value)
    return None


def _extract_values(candidate: Any) -> List[Any]:
    """Normalise different iterable payloads into a flat list of values."""

    if candidate is None:
        return []
    if isinstance(candidate, Mapping):
        values: List[Any] = []
        for key in ("new", "new_values", "unexpected", "unexpected_values", "values", "items"):
            inner = candidate.get(key)
            if isinstance(inner, (list, tuple, set)):
                values.extend(inner)
            elif inner is not None:
                values.append(inner)
        return values
    if isinstance(candidate, (list, tuple, set)):
        return list(candidate)
    return [candidate]


def _enum_extension(
    *,
    dq: DataQuality,
    metrics: Mapping[str, Any],
    field_name: str,
) -> Optional[Tuple[List[Any], List[Any]]]:
    """Return updated enum values plus additions derived from observations."""

    if not field_name:
        return None
    base_values: List[Any]
    if isinstance(dq.mustBe, (list, tuple, set)):
        base_values = list(dq.mustBe)
    else:
        return None

    observed_sources = [
        metrics.get(f"observed.enum_{field_name}"),
        metrics.get("observed.enum", {}),
    ]
    observed_values: List[Any] = []
    for source in observed_sources:
        if isinstance(source, Mapping) and field_name in source:
            observed_values.extend(_extract_values(source.get(field_name)))
        else:
            observed_values.extend(_extract_values(source))

    if not observed_values:
        return None

    seen = {str(v) for v in base_values}
    additions: List[Any] = []
    for value in observed_values:
        key = str(value)
        if key not in seen:
            additions.append(value)
            seen.add(key)

    if not additions:
        return None

    updated = list(base_values) + additions

    return updated, additions


def draft_from_validation_result(
    *,
    validation: ValidationResult,
    base_contract: OpenDataContractStandard,
    bump: str = "minor",
    dataset_id: Optional[str] = None,
    dataset_version: Optional[str] = None,
    data_format: Optional[str] = None,
    dq_status: Optional[str] = None,
    dq_feedback: Optional[Mapping[str, Any]] = None,
    draft_context: Optional[Mapping[str, Any]] = None,
) -> Optional[OpenDataContractStandard]:
    """Return a draft contract derived from validation feedback."""

    metrics = validation.metrics or {}
    schema = validation.schema or {}

    has_errors = bool(validation.errors)
    has_warnings = bool(validation.warnings)
    if not has_errors and not has_warnings:
        return None

    contract_id, version = contract_identity(base_contract)
    bump_version = SemVer.parse(version).bump(bump)

    if hasattr(base_contract, "model_copy"):
        draft = base_contract.model_copy(deep=True)  # type: ignore[attr-defined]
    else:
        draft = to_model(as_odcs_dict(base_contract))
    draft.version = str(bump_version)
    draft.status = "draft"

    suffix = _draft_version_suffix(
        dataset_id=dataset_id,
        dataset_version=dataset_version,
        draft_context=draft_context,
    )
    draft.version = f"{draft.version}-{suffix}"

    context_payload: Dict[str, Any] = dict(draft_context or {})
    if dataset_id and "dataset_id" not in context_payload:
        context_payload["dataset_id"] = dataset_id
    if dataset_version and "dataset_version" not in context_payload:
        context_payload["dataset_version"] = dataset_version

    pipeline_token = _pipeline_hint(draft_context)
    pipeline_value: Optional[str] = None
    if draft_context:
        for key in ("pipeline", "job", "project", "module", "function", "qualname", "source"):
            raw = draft_context.get(key)
            if raw:
                pipeline_value = str(raw)
                break

    change_log: List[Dict[str, Any]] = []
    change_log = _apply_schema_feedback(
        draft,
        schema=schema,
        metrics=metrics,
        change_log=change_log,
    )

    if validation.errors:
        change_log.append(
            {
                "status": "error",
                "kind": "validation",
                "messages": list(validation.errors),
            }
        )
    if validation.warnings:
        change_log.append(
            {
                "status": "warning",
                "kind": "validation",
                "messages": list(validation.warnings),
            }
        )

    try:
        draft_custom_properties = draft.customProperties
    except AttributeError:
        draft_custom_properties = None
    custom_properties = list(normalise_custom_properties(draft_custom_properties))

    if dq_status or dq_feedback:
        feedback = dict(dq_feedback or {})
        if dq_status:
            feedback.setdefault("status", dq_status)
        custom_properties.append(CustomProperty(property="dq_feedback", value=feedback))

    custom_properties.append(
        CustomProperty(
            property="validation_metrics",
            value={"metrics": metrics, "schema": schema},
        )
    )

    if data_format:
        custom_properties.append(CustomProperty(property="data_format", value=data_format))

    custom_properties.append(
        CustomProperty(
            property="base_contract",
            value={"id": contract_id, "version": version},
        )
    )

    custom_properties.append(
        CustomProperty(
            property="validation_outcome",
            value={"errors": validation.errors, "warnings": validation.warnings},
        )
    )

    if context_payload:
        if pipeline_value and "module" not in context_payload:
            module_hint = pipeline_value.rsplit(".", 1)[0]
            context_payload.setdefault("module", module_hint)
        custom_properties.append(
            CustomProperty(property="draft_context", value=context_payload)
        )

    if pipeline_value:
        custom_properties.append(
            CustomProperty(property="draft_pipeline", value=pipeline_value)
        )
    elif pipeline_token:
        custom_properties.append(
            CustomProperty(property="draft_pipeline", value=pipeline_token)
        )

    provenance: Dict[str, Any] = {}
    if dataset_version:
        provenance["dataset_version"] = dataset_version
    if dataset_id:
        provenance["dataset_id"] = dataset_id
    if provenance:
        custom_properties.append(
            CustomProperty(property="provenance", value=provenance)
        )

    if dataset_id or dataset_version:
        reference = {
            "dataset_id": dataset_id,
            "dataset_version": dataset_version,
            "collected_at": datetime.now(timezone.utc).isoformat(),
        }
        custom_properties.append(
            CustomProperty(property="validation_reference", value=reference)
        )

    custom_properties.append(
        CustomProperty(property="draft_change_log", value=change_log)
    )

    draft.customProperties = custom_properties

    return draft


def draft_from_observations(
    *,
    observations: Mapping[str, Mapping[str, Any]] | None,
    base_contract: OpenDataContractStandard,
    dataset_id: Optional[str] = None,
    dataset_version: Optional[str] = None,
    draft_context: Optional[Mapping[str, Any]] = None,
) -> OpenDataContractStandard:
    """Return a draft contract using observed schema information only."""

    if hasattr(base_contract, "model_copy"):
        draft = base_contract.model_copy(deep=True)  # type: ignore[attr-defined]
    else:
        draft = to_model(as_odcs_dict(base_contract))
    contract_id, version = contract_identity(base_contract)
    bump_version = SemVer.parse(version).bump("patch")

    suffix = _draft_version_suffix(
        dataset_id=dataset_id,
        dataset_version=dataset_version,
        draft_context=draft_context,
    )
    draft.version = f"{bump_version}-{suffix}"
    draft.status = "draft"

    context_payload: Dict[str, Any] = dict(draft_context or {})
    if dataset_id and "dataset_id" not in context_payload:
        context_payload["dataset_id"] = dataset_id
    if dataset_version and "dataset_version" not in context_payload:
        context_payload["dataset_version"] = dataset_version

    pipeline_token = _pipeline_hint(draft_context)
    pipeline_value: Optional[str] = None
    if draft_context:
        for key in ("pipeline", "job", "project", "module", "function", "qualname", "source"):
            raw = draft_context.get(key)
            if raw:
                pipeline_value = str(raw)
                break

    change_log = _apply_schema_feedback(
        draft,
        schema=observations or {},
        metrics={},
        change_log=[],
    )

    try:
        draft_custom_properties = draft.customProperties
    except AttributeError:
        draft_custom_properties = None
    custom_properties = list(normalise_custom_properties(draft_custom_properties))
    custom_properties.append(
        CustomProperty(
            property="base_contract",
            value={"id": contract_id, "version": version},
        )
    )
    custom_properties.append(
        CustomProperty(property="observed_schema", value=observations or {})
    )

    if context_payload:
        if pipeline_value and "module" not in context_payload:
            module_hint = pipeline_value.rsplit(".", 1)[0]
            context_payload.setdefault("module", module_hint)
        custom_properties.append(
            CustomProperty(property="draft_context", value=context_payload)
        )
    if pipeline_value:
        custom_properties.append(
            CustomProperty(property="draft_pipeline", value=pipeline_value)
        )
    elif pipeline_token:
        custom_properties.append(
            CustomProperty(property="draft_pipeline", value=pipeline_token)
        )
    custom_properties.append(
        CustomProperty(property="draft_change_log", value=change_log)
    )

    draft.customProperties = custom_properties

    return draft


def _apply_schema_feedback(
    draft: OpenDataContractStandard,
    *,
    schema: Mapping[str, Mapping[str, Any]],
    metrics: Mapping[str, Any],
    change_log: Optional[List[Dict[str, Any]]] = None,
) -> List[Dict[str, Any]]:
    """Update ``draft`` schema using observed field metadata."""

    log: List[Dict[str, Any]] = change_log if change_log is not None else []

    for obj in draft.schema_ or []:
        for field in obj.properties or []:
            name = field.name
            if not name:
                continue
            observed = schema.get(name) or {}
            observed_type, nullable = _resolve_observed_type(
                observed,
                field.physicalType or field.logicalType,
            )
            if observed_type:
                field.physicalType = observed_type
            was_required = bool(field.required)
            if nullable is not None:
                field.required = not nullable
            if was_required and not field.required:
                log.append({
                    "field": name,
                    "status": "relaxed",
                    "constraint": "required",
                })
            if observed:
                field.description = field.description or ""
                field.description = (
                    f"{field.description}\nObserved metadata: {observed}".strip()
                )

            updated_quality: List[DataQuality] = []
            for dq in list(field.quality or []):
                result = _quality_rule_key(field, dq)
                if not result:
                    updated_quality.append(dq)
                    continue
                prefix, label = result
                value = _quality_metric_value(
                    metrics=metrics,
                    rule_prefix=prefix,
                    field_name=name,
                )
                if prefix == "enum":
                    extension = _enum_extension(dq=dq, metrics=metrics, field_name=name)
                    if extension:
                        updated, additions = extension
                        dq.mustBe = updated
                        log.append({
                            "field": name,
                            "rule": "enum",
                            "status": "updated",
                            "details": {"added_values": additions},
                        })
                    else:
                        log.append({
                            "field": name,
                            "rule": "enum",
                            "status": "kept",
                        })
                    updated_quality.append(dq)
                    continue

                if value and value > 0:
                    log.append({
                        "field": name,
                        "rule": label,
                        "status": "removed",
                        "details": {"violations": value},
                    })
                    continue

                log.append({
                    "field": name,
                    "rule": label,
                    "status": "kept",
                })
                dq.description = dq.description or ""
                dq.description = (
                    f"{dq.description}\nObserved {label}: {value}".strip()
                )
                updated_quality.append(dq)

            field.quality = updated_quality or None

    return log


__all__ = [
    "draft_from_observations",
    "draft_from_validation_result",
]



"""Delta-backed data product backend for Unity Catalog deployments."""

from __future__ import annotations

try:  # pragma: no cover - optional dependency
    from pyspark.sql import SparkSession
except Exception:  # pragma: no cover - Spark is optional for most deployments
    SparkSession = object  # type: ignore

from .local import _StoreBackedDataProductServiceBackend
from .stores import DataProductStore

try:  # pragma: no cover - optional dependency
    from .stores.delta import DeltaDataProductStore
except ModuleNotFoundError:  # pragma: no cover - pyspark optional
    DeltaDataProductStore = None  # type: ignore[assignment]


class DeltaDataProductServiceBackend(_StoreBackedDataProductServiceBackend):
    """Persist ODPS documents in a Delta table or Unity Catalog object."""

    def __init__(
        self,
        spark: SparkSession,
        *,
        table: str | None = None,
        path: str | None = None,
        store: DataProductStore | None = None,
        log_sql: bool = False,
    ) -> None:
        if store is None:
            if DeltaDataProductStore is None:
                raise RuntimeError(
                    "pyspark is required when using the Delta data product backend"
                )
            store = DeltaDataProductStore(spark, table=table, path=path, log_sql=log_sql)
        super().__init__(store)


__all__ = ["DeltaDataProductServiceBackend"]



"""SQL-backed contract store implementation."""

from __future__ import annotations

from typing import List
import json

from sqlalchemy import (
    Column,
    DateTime,
    MetaData,
    String,
    Table,
    Text,
    and_,
    create_engine,
    func,
    insert,
    select,
    update,
)
from sqlalchemy.engine import Engine

from ._sql_common import prepare_contract_row
from .interface import ContractStore
from dc43_service_backends.core.odcs import to_model
from open_data_contract_standard.model import OpenDataContractStandard  # type: ignore


class SQLContractStore(ContractStore):
    """Persist contracts in a relational database using SQLAlchemy Core."""

    def __init__(self, engine: Engine, table_name: str = "contracts", schema: str | None = None):
        self.engine = engine
        self.metadata = MetaData(schema=schema)
        self.table = Table(
            table_name,
            self.metadata,
            Column("contract_id", String(255), primary_key=True, nullable=False),
            Column("version", String(64), primary_key=True, nullable=False),
            Column("name", String(255), nullable=False),
            Column("description", Text),
            Column("json", Text, nullable=False),
            Column("fingerprint", String(64), nullable=False),
            Column("created_at", DateTime(timezone=True), nullable=False, server_default=func.now()),
        )
        self.metadata.create_all(self.engine)

    def put(self, contract: OpenDataContractStandard) -> None:
        """Insert or update a contract document."""

        cid, ver, payload = prepare_contract_row(contract)

        with self.engine.begin() as conn:
            stmt = select(self.table.c.contract_id).where(
                and_(
                    self.table.c.contract_id == cid,
                    self.table.c.version == ver,
                )
            )
            exists = conn.execute(stmt).first() is not None
            if exists:
                conn.execute(
                    update(self.table)
                    .where(
                        and_(
                            self.table.c.contract_id == cid,
                            self.table.c.version == ver,
                        )
                    )
                    .values(payload)
                )
            else:
                conn.execute(insert(self.table).values(payload))

    def get(self, contract_id: str, version: str) -> OpenDataContractStandard:
        """Fetch a contract from the relational table and return it as a model."""

        with self.engine.begin() as conn:
            stmt = select(self.table.c.json).where(
                and_(
                    self.table.c.contract_id == contract_id,
                    self.table.c.version == version,
                )
            )
            row = conn.execute(stmt).first()

        if not row:
            raise KeyError(f"Contract {contract_id}:{version} not found")

        return to_model(json.loads(row[0]))

    def list_contracts(self) -> List[str]:
        with self.engine.begin() as conn:
            stmt = select(self.table.c.contract_id).distinct().order_by(self.table.c.contract_id)
            rows = conn.execute(stmt).scalars().all()
        return list(rows)

    def list_versions(self, contract_id: str) -> List[str]:
        with self.engine.begin() as conn:
            stmt = (
                select(self.table.c.version)
                .where(self.table.c.contract_id == contract_id)
                .order_by(self.table.c.version)
            )
            rows = conn.execute(stmt).scalars().all()
        return list(rows)

__all__ = ["SQLContractStore", "create_engine"]



from __future__ import annotations

"""Configuration helpers for the dc43 service backend HTTP application."""

from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Iterable, Mapping, MutableMapping, Sequence
import json
import os
import re

import tomllib

try:
    import tomlkit
except ModuleNotFoundError:  # pragma: no cover - exercised via fallback tests
    tomlkit = None

__all__ = [
    "ContractStoreConfig",
    "DataProductStoreConfig",
    "DataQualityBackendConfig",
    "AuthConfig",
    "GovernanceConfig",
    "GovernanceStoreConfig",
    "UnityCatalogConfig",
    "ServiceBackendsConfig",
    "load_config",
    "config_to_mapping",
    "dumps",
    "dump",
]


_BARE_KEY_PATTERN = re.compile(r"^[A-Za-z0-9_-]+$")


def _format_key(value: str) -> str:
    """Return ``value`` formatted as a TOML key."""

    if _BARE_KEY_PATTERN.match(value):
        return value
    return json.dumps(value)


def _format_value(value: Any) -> str:
    """Return ``value`` rendered as TOML without relying on ``tomlkit``."""

    if isinstance(value, str):
        return json.dumps(value)
    if isinstance(value, bool):
        return "true" if value else "false"
    if isinstance(value, int):
        return str(value)
    if isinstance(value, float):
        return repr(value)
    if value is None:
        return '""'
    if isinstance(value, Mapping):
        if not value:
            return "{}"
        items = [
            f"{_format_key(str(key))} = {_format_value(item)}"
            for key, item in value.items()
        ]
        return "{ " + ", ".join(items) + " }"
    if isinstance(value, Sequence) and not isinstance(value, (bytes, bytearray, str)):
        values = ", ".join(_format_value(item) for item in value)
        return f"[ {values} ]" if values else "[]"
    return json.dumps(str(value))


def _join_table(parts: Iterable[str]) -> str:
    """Return the TOML dotted path for ``parts``."""

    return ".".join(_format_key(part) for part in parts)


def _write_table(
    mapping: Mapping[str, Any],
    lines: list[str],
    prefix: tuple[str, ...] = (),
) -> None:
    """Append TOML lines representing ``mapping`` to ``lines``."""

    scalar_items: list[tuple[str, Any]] = []
    table_items: list[tuple[str, Mapping[str, Any]]] = []

    for key, value in mapping.items():
        key_str = str(key)
        if isinstance(value, Mapping):
            table_items.append((key_str, value))
            continue
        scalar_items.append((key_str, value))

    for key, value in scalar_items:
        lines.append(f"{_format_key(key)} = {_format_value(value)}")

    for key, value in table_items:
        table_prefix = prefix + (key,)
        has_scalars = any(not isinstance(item, Mapping) for item in value.values())
        if has_scalars or not value:
            if lines and lines[-1] != "":
                lines.append("")
            lines.append(f"[{_join_table(table_prefix)}]")
        _write_table(value, lines, table_prefix)


def _toml_dumps(payload: Mapping[str, Any]) -> str:
    """Return TOML for ``payload`` using ``tomlkit`` when available."""

    if tomlkit is not None:  # pragma: no branch
        return tomlkit.dumps(payload)
    lines: list[str] = []
    _write_table(payload, lines)
    if not lines:
        return ""
    text = "\n".join(lines)
    if not text.endswith("\n"):
        text += "\n"
    return text


@dataclass(slots=True)
class ContractStoreConfig:
    """Configuration for the active contract store implementation."""

    type: str = "filesystem"
    root: Path | None = None
    base_path: Path | None = None
    table: str | None = None
    base_url: str | None = None
    dsn: str | None = None
    schema: str | None = None
    token: str | None = None
    timeout: float = 10.0
    contracts_endpoint_template: str | None = None
    default_status: str = "Draft"
    status_filter: str | None = None
    catalog: dict[str, tuple[str, str]] = field(default_factory=dict)
    log_sql: bool = False


@dataclass(slots=True)
class AuthConfig:
    """Authentication configuration for protecting backend endpoints."""

    token: str | None = None


@dataclass(slots=True)
class DataProductStoreConfig:
    """Configuration for the active data product store implementation."""

    type: str = "memory"
    root: Path | None = None
    base_path: Path | None = None
    table: str | None = None
    dsn: str | None = None
    schema: str | None = None
    base_url: str | None = None
    catalog: str | None = None
    log_sql: bool = False


@dataclass(slots=True)
class DataQualityBackendConfig:
    """Configuration for data-quality backend delegates."""

    type: str = "local"
    base_url: str | None = None
    token: str | None = None
    token_header: str = "Authorization"
    token_scheme: str = "Bearer"
    headers: dict[str, str] = field(default_factory=dict)
    default_engine: str = "native"
    engines: dict[str, dict[str, object]] = field(default_factory=dict)


@dataclass(slots=True)
class GovernanceStoreConfig:
    """Configuration for persisting governance artefacts."""

    type: str = "memory"
    root: Path | None = None
    base_path: Path | None = None
    table: str | None = None
    status_table: str | None = None
    activity_table: str | None = None
    link_table: str | None = None
    metrics_table: str | None = None
    dsn: str | None = None
    schema: str | None = None
    base_url: str | None = None
    token: str | None = None
    token_header: str = "Authorization"
    token_scheme: str = "Bearer"
    timeout: float = 10.0
    headers: dict[str, str] = field(default_factory=dict)
    log_sql: bool = False


@dataclass(slots=True)
class UnityCatalogConfig:
    """Optional Databricks Unity Catalog synchronisation settings."""

    enabled: bool = False
    dataset_prefix: str = "table:"
    workspace_profile: str | None = None
    workspace_url: str | None = None
    workspace_token: str | None = None
    sql_dsn: str | None = None
    tags_enabled: bool = False
    tags_sql_dsn: str | None = None
    static_properties: dict[str, str] = field(default_factory=dict)
    static_tags: dict[str, str] = field(default_factory=dict)

    @property
    def workspace_host(self) -> str | None:
        """Backwards-compatible accessor for legacy ``workspace_host`` keys."""

        return self.workspace_url

    @workspace_host.setter
    def workspace_host(self, value: str | None) -> None:
        self.workspace_url = value


@dataclass(slots=True)
class GovernanceConfig:
    """Governance service extension wiring sourced from configuration."""

    dataset_contract_link_builders: tuple[str, ...] = field(default_factory=tuple)


@dataclass(slots=True)
class ServiceBackendsConfig:
    """Top level configuration for the service backend application."""

    contract_store: ContractStoreConfig = field(default_factory=ContractStoreConfig)
    data_product_store: DataProductStoreConfig = field(default_factory=DataProductStoreConfig)
    data_quality: DataQualityBackendConfig = field(default_factory=DataQualityBackendConfig)
    auth: AuthConfig = field(default_factory=AuthConfig)
    unity_catalog: UnityCatalogConfig = field(default_factory=UnityCatalogConfig)
    governance: GovernanceConfig = field(default_factory=GovernanceConfig)
    governance_store: GovernanceStoreConfig = field(default_factory=GovernanceStoreConfig)


def _first_existing_path(paths: list[str | os.PathLike[str] | None]) -> Path | None:
    for candidate in paths:
        if not candidate:
            continue
        resolved = Path(candidate).expanduser()
        if resolved.is_file():
            return resolved
    return None


def _load_toml(path: Path | None) -> Mapping[str, Any]:
    if not path:
        return {}
    try:
        data = path.read_bytes()
    except OSError:
        return {}
    try:
        return tomllib.loads(data.decode("utf-8"))
    except tomllib.TOMLDecodeError:
        return {}


def _coerce_path(value: Any) -> Path | None:
    if value in {None, ""}:
        return None
    return Path(str(value)).expanduser()


def _coerce_float(value: Any, default: float) -> float:
    try:
        return float(value)
    except (TypeError, ValueError):
        return default


def _parse_catalog(section: Any) -> dict[str, tuple[str, str]]:
    catalog: dict[str, tuple[str, str]] = {}
    if not isinstance(section, MutableMapping):
        return catalog
    for contract_id, mapping in section.items():
        if not isinstance(mapping, MutableMapping):
            continue
        data_product = mapping.get("data_product")
        port = mapping.get("port")
        if data_product is None or port is None:
            continue
        contract_key = str(contract_id).strip()
        if not contract_key:
            continue
        catalog[contract_key] = (str(data_product).strip(), str(port).strip())
    return catalog


def _parse_bool(value: Any, default: bool) -> bool:
    if isinstance(value, bool):
        return value
    if isinstance(value, (int, float)):
        return bool(value)
    if isinstance(value, str):
        normalised = value.strip().lower()
        if normalised in {"1", "true", "yes", "on"}:
            return True
        if normalised in {"0", "false", "no", "off"}:
            return False
    return default


def _parse_str_dict(section: Any) -> dict[str, str]:
    values: dict[str, str] = {}
    if not isinstance(section, MutableMapping):
        return values
    for key, value in section.items():
        key_str = str(key).strip()
        if not key_str:
            continue
        values[key_str] = str(value)
    return values


def load_config(path: str | os.PathLike[str] | None = None) -> ServiceBackendsConfig:
    """Load configuration from ``path`` or fall back to defaults."""

    default_path = Path(__file__).with_name("config").joinpath("default.toml")
    env_path = os.getenv("DC43_SERVICE_BACKENDS_CONFIG")
    config_path = _first_existing_path([path, env_path, default_path])
    payload = _load_toml(config_path)

    store_section = (
        payload.get("contract_store")
        if isinstance(payload, MutableMapping)
        else {}
    )
    data_product_section = (
        payload.get("data_product")
        if isinstance(payload, MutableMapping)
        else {}
    )

    auth_section = (
        payload.get("auth")
        if isinstance(payload, MutableMapping)
        else {}
    )
    dq_section = (
        payload.get("data_quality")
        if isinstance(payload, MutableMapping)
        else {}
    )

    governance_store_section = (
        payload.get("governance_store")
        if isinstance(payload, MutableMapping)
        else {}
    )

    unity_section = (
        payload.get("unity_catalog")
        if isinstance(payload, MutableMapping)
        else {}
    )
    governance_section = (
        payload.get("governance")
        if isinstance(payload, MutableMapping)
        else {}
    )

    store_type = "filesystem"
    root_value = None
    base_path_value = None
    table_value = None
    base_url_value = None
    dsn_value = None
    schema_value = None
    store_token_value = None
    timeout_value = 10.0
    endpoint_template = None
    default_status = "Draft"
    status_filter = None
    catalog_value: dict[str, tuple[str, str]] = {}
    store_log_sql = False
    if isinstance(store_section, MutableMapping):
        raw_type = store_section.get("type")
        if isinstance(raw_type, str) and raw_type.strip():
            store_type = raw_type.strip().lower()
        root_value = _coerce_path(store_section.get("root"))
        base_path_value = _coerce_path(store_section.get("base_path"))
        table_raw = store_section.get("table")
        if isinstance(table_raw, str) and table_raw.strip():
            table_value = table_raw.strip()
        base_url_raw = store_section.get("base_url")
        if base_url_raw is not None:
            base_url_value = str(base_url_raw).strip() or None
        dsn_raw = store_section.get("dsn")
        if dsn_raw is not None:
            dsn_value = str(dsn_raw).strip() or None
        schema_raw = store_section.get("schema")
        if schema_raw is not None:
            schema_value = str(schema_raw).strip() or None
        token_raw = store_section.get("token")
        if token_raw is not None:
            store_token_value = str(token_raw).strip() or None
        timeout_value = _coerce_float(store_section.get("timeout"), 10.0)
        template_raw = store_section.get("contracts_endpoint_template")
        if template_raw is not None:
            endpoint_template = str(template_raw).strip() or None
        default_status = str(store_section.get("default_status", "Draft")).strip() or "Draft"
        status_raw = store_section.get("status_filter")
        if status_raw is not None:
            status_filter = str(status_raw).strip() or None
        catalog_value = _parse_catalog(store_section.get("catalog"))
        store_log_sql = _parse_bool(store_section.get("log_sql"), False)

    dp_type = "memory"
    dp_root_value = None
    dp_base_path_value = None
    dp_table_value = None
    dp_dsn_value = None
    dp_schema_value = None
    dp_base_url_value = None
    dp_catalog_value = None
    dp_log_sql = False
    if isinstance(data_product_section, MutableMapping):
        raw_type = data_product_section.get("type")
        if isinstance(raw_type, str) and raw_type.strip():
            dp_type = raw_type.strip().lower()
        dp_root_value = _coerce_path(data_product_section.get("root"))
        dp_base_path_value = _coerce_path(data_product_section.get("base_path"))
        table_raw = data_product_section.get("table")
        if isinstance(table_raw, str) and table_raw.strip():
            dp_table_value = table_raw.strip()
        dsn_raw = data_product_section.get("dsn")
        if dsn_raw is not None:
            dp_dsn_value = str(dsn_raw).strip() or None
        schema_raw = data_product_section.get("schema")
        if schema_raw is not None:
            dp_schema_value = str(schema_raw).strip() or None
        base_url_raw = data_product_section.get("base_url")
        if base_url_raw is not None:
            dp_base_url_value = str(base_url_raw).strip() or None
        catalog_raw = data_product_section.get("catalog")
        if catalog_raw is not None:
            dp_catalog_value = str(catalog_raw).strip() or None
        dp_log_sql = _parse_bool(data_product_section.get("log_sql"), False)

    dq_type = "local"
    dq_base_url_value = None
    dq_token_value = None
    dq_token_header_value = "Authorization"
    dq_token_scheme_value = "Bearer"
    dq_headers_value: dict[str, str] = {}
    dq_default_engine_value = "native"
    dq_engines_value: dict[str, dict[str, object]] = {}
    if isinstance(dq_section, MutableMapping):
        raw_type = dq_section.get("type")
        if isinstance(raw_type, str) and raw_type.strip():
            dq_type = raw_type.strip().lower()
        base_url_raw = dq_section.get("base_url")
        if base_url_raw is not None:
            dq_base_url_value = str(base_url_raw).strip() or None
        token_raw = dq_section.get("token")
        if token_raw is not None:
            dq_token_value = str(token_raw).strip() or None
        token_header_raw = dq_section.get("token_header")
        if token_header_raw is not None:
            dq_token_header_value = str(token_header_raw).strip()
        token_scheme_raw = dq_section.get("token_scheme")
        if token_scheme_raw is not None:
            dq_token_scheme_value = str(token_scheme_raw).strip()
        headers_raw = dq_section.get("headers")
        if headers_raw is not None:
            dq_headers_value = _parse_str_dict(headers_raw)
        default_engine_raw = dq_section.get("default_engine")
        if isinstance(default_engine_raw, str) and default_engine_raw.strip():
            dq_default_engine_value = default_engine_raw.strip()
        engines_raw = dq_section.get("engines")
        if isinstance(engines_raw, MutableMapping):
            engine_mapping: dict[str, dict[str, object]] = {}
            for name, engine_section in engines_raw.items():
                key = str(name).strip()
                if not key or not isinstance(engine_section, MutableMapping):
                    continue
                engine_mapping[key] = {str(k): v for k, v in engine_section.items()}
            dq_engines_value = engine_mapping

    auth_token_value = None
    if isinstance(auth_section, MutableMapping):
        token_raw = auth_section.get("token")
        if token_raw is not None:
            auth_token_value = str(token_raw).strip() or None

    unity_enabled = False
    unity_prefix = "table:"
    unity_profile = None
    unity_url = None
    unity_token = None
    unity_sql_dsn = None
    unity_tags_enabled = False
    unity_tags_sql_dsn = None
    unity_static: dict[str, str] = {}
    unity_static_tags: dict[str, str] = {}
    if isinstance(unity_section, MutableMapping):
        unity_enabled = _parse_bool(unity_section.get("enabled"), False)
        prefix_raw = unity_section.get("dataset_prefix")
        if isinstance(prefix_raw, str) and prefix_raw.strip():
            unity_prefix = prefix_raw.strip()
        profile_raw = unity_section.get("workspace_profile")
        if isinstance(profile_raw, str) and profile_raw.strip():
            unity_profile = profile_raw.strip()
        host_raw = unity_section.get("workspace_url")
        if host_raw is None:
            host_raw = unity_section.get("workspace_host")
        if isinstance(host_raw, str) and host_raw.strip():
            unity_url = host_raw.strip()
        token_raw = unity_section.get("workspace_token")
        if isinstance(token_raw, str) and token_raw.strip():
            unity_token = token_raw.strip()
        sql_dsn_raw = unity_section.get("sql_dsn")
        if isinstance(sql_dsn_raw, str) and sql_dsn_raw.strip():
            unity_sql_dsn = sql_dsn_raw.strip()
        tags_enabled_raw = unity_section.get("tags_enabled")
        unity_tags_enabled = _parse_bool(tags_enabled_raw, False)
        tags_sql_dsn_raw = unity_section.get("tags_sql_dsn")
        if isinstance(tags_sql_dsn_raw, str) and tags_sql_dsn_raw.strip():
            unity_tags_sql_dsn = tags_sql_dsn_raw.strip()
        unity_static = _parse_str_dict(unity_section.get("static_properties"))
        unity_static_tags = _parse_str_dict(unity_section.get("static_tags"))

    link_builder_specs: list[str] = []
    if isinstance(governance_section, MutableMapping):
        raw_builders = governance_section.get("dataset_contract_link_builders")
        if isinstance(raw_builders, (list, tuple, set)):
            for entry in raw_builders:
                text = str(entry).strip()
                if text:
                    link_builder_specs.append(text)
        elif isinstance(raw_builders, str):
            for chunk in raw_builders.split(","):
                text = chunk.strip()
                if text:
                    link_builder_specs.append(text)

    gov_store_type = "memory"
    gov_root_value = None
    gov_base_path_value = None
    gov_table_value = None
    gov_status_table_value = None
    gov_activity_table_value = None
    gov_link_table_value = None
    gov_dsn_value = None
    gov_schema_value = None
    gov_base_url_value = None
    gov_token_value = None
    gov_token_header_value = "Authorization"
    gov_token_scheme_value = "Bearer"
    gov_timeout_value = 10.0
    gov_headers_value: dict[str, str] = {}
    gov_log_sql = False
    gov_metrics_table_value = None
    if isinstance(governance_store_section, MutableMapping):
        raw_type = governance_store_section.get("type")
        if isinstance(raw_type, str) and raw_type.strip():
            gov_store_type = raw_type.strip().lower()
        gov_root_value = _coerce_path(governance_store_section.get("root"))
        gov_base_path_value = _coerce_path(governance_store_section.get("base_path"))
        table_raw = governance_store_section.get("table")
        if isinstance(table_raw, str) and table_raw.strip():
            gov_table_value = table_raw.strip()
        status_table_raw = governance_store_section.get("status_table")
        if isinstance(status_table_raw, str) and status_table_raw.strip():
            gov_status_table_value = status_table_raw.strip()
        activity_table_raw = governance_store_section.get("activity_table")
        if isinstance(activity_table_raw, str) and activity_table_raw.strip():
            gov_activity_table_value = activity_table_raw.strip()
        link_table_raw = governance_store_section.get("link_table")
        if isinstance(link_table_raw, str) and link_table_raw.strip():
            gov_link_table_value = link_table_raw.strip()
        metrics_table_raw = governance_store_section.get("metrics_table")
        if isinstance(metrics_table_raw, str) and metrics_table_raw.strip():
            gov_metrics_table_value = metrics_table_raw.strip()
        dsn_raw = governance_store_section.get("dsn")
        if dsn_raw is not None:
            gov_dsn_value = str(dsn_raw).strip() or None
        schema_raw = governance_store_section.get("schema")
        if schema_raw is not None:
            gov_schema_value = str(schema_raw).strip() or None
        base_url_raw = governance_store_section.get("base_url")
        if base_url_raw is not None:
            gov_base_url_value = str(base_url_raw).strip() or None
        token_raw = governance_store_section.get("token")
        if token_raw is not None:
            gov_token_value = str(token_raw).strip() or None
        token_header_raw = governance_store_section.get("token_header")
        if token_header_raw is not None:
            gov_token_header_value = str(token_header_raw).strip() or "Authorization"
        token_scheme_raw = governance_store_section.get("token_scheme")
        if token_scheme_raw is not None:
            gov_token_scheme_value = str(token_scheme_raw).strip() or "Bearer"
        gov_timeout_value = _coerce_float(governance_store_section.get("timeout"), 10.0)
        headers_raw = governance_store_section.get("headers")
        if headers_raw is not None:
            gov_headers_value = _parse_str_dict(headers_raw)
        gov_log_sql = _parse_bool(governance_store_section.get("log_sql"), False)

    env_contract_type = os.getenv("DC43_CONTRACT_STORE_TYPE")
    if env_contract_type:
        normalised_type = env_contract_type.strip().lower()
        if normalised_type:
            store_type = normalised_type

    env_root = os.getenv("DC43_CONTRACT_STORE")
    if env_root:
        root_value = _coerce_path(env_root)
        base_path_value = root_value if base_path_value is None else base_path_value

    env_contract_table = os.getenv("DC43_CONTRACT_STORE_TABLE")
    if env_contract_table:
        table_value = env_contract_table.strip() or table_value

    env_contract_dsn = os.getenv("DC43_CONTRACT_STORE_DSN")
    if env_contract_dsn:
        dsn_value = env_contract_dsn.strip() or dsn_value

    env_contract_schema = os.getenv("DC43_CONTRACT_STORE_SCHEMA")
    if env_contract_schema:
        schema_value = env_contract_schema.strip() or schema_value

    env_contract_log_sql = os.getenv("DC43_CONTRACT_STORE_LOG_SQL")
    if env_contract_log_sql is not None:
        store_log_sql = _parse_bool(env_contract_log_sql, store_log_sql)

    env_dp_root = os.getenv("DC43_DATA_PRODUCT_STORE")
    if env_dp_root:
        dp_root_value = _coerce_path(env_dp_root)
        dp_base_path_value = dp_root_value if dp_base_path_value is None else dp_base_path_value

    env_dp_table = os.getenv("DC43_DATA_PRODUCT_TABLE")
    if env_dp_table:
        dp_table_value = env_dp_table.strip() or dp_table_value

    env_dp_dsn = os.getenv("DC43_DATA_PRODUCT_STORE_DSN")
    if env_dp_dsn:
        dp_dsn_value = env_dp_dsn.strip() or dp_dsn_value

    env_dp_schema = os.getenv("DC43_DATA_PRODUCT_STORE_SCHEMA")
    if env_dp_schema:
        dp_schema_value = env_dp_schema.strip() or dp_schema_value

    env_dp_log_sql = os.getenv("DC43_DATA_PRODUCT_STORE_LOG_SQL")
    if env_dp_log_sql is not None:
        dp_log_sql = _parse_bool(env_dp_log_sql, dp_log_sql)

    env_token = os.getenv("DC43_BACKEND_TOKEN")
    if env_token:
        auth_token_value = env_token.strip() or None

    env_dq_type = os.getenv("DC43_DATA_QUALITY_BACKEND_TYPE")
    if env_dq_type:
        normalised = env_dq_type.strip().lower()
        if normalised:
            dq_type = normalised

    env_dq_url = os.getenv("DC43_DATA_QUALITY_BACKEND_URL")
    if env_dq_url:
        dq_base_url_value = env_dq_url.strip() or dq_base_url_value

    env_dq_token = os.getenv("DC43_DATA_QUALITY_BACKEND_TOKEN")
    if env_dq_token:
        dq_token_value = env_dq_token.strip() or dq_token_value

    env_dq_header = os.getenv("DC43_DATA_QUALITY_BACKEND_TOKEN_HEADER")
    if env_dq_header is not None:
        dq_token_header_value = env_dq_header.strip()

    env_dq_scheme = os.getenv("DC43_DATA_QUALITY_BACKEND_TOKEN_SCHEME")
    if env_dq_scheme is not None:
        dq_token_scheme_value = env_dq_scheme.strip()

    env_dq_default_engine = os.getenv("DC43_DATA_QUALITY_DEFAULT_ENGINE")
    if env_dq_default_engine:
        dq_default_engine_value = env_dq_default_engine.strip() or dq_default_engine_value

    env_unity_enabled = os.getenv("DC43_UNITY_CATALOG_ENABLED")
    if env_unity_enabled is not None:
        unity_enabled = _parse_bool(env_unity_enabled, unity_enabled)

    env_unity_prefix = os.getenv("DC43_UNITY_CATALOG_PREFIX")
    if env_unity_prefix:
        unity_prefix = env_unity_prefix.strip() or unity_prefix

    env_profile = os.getenv("DATABRICKS_CONFIG_PROFILE")
    if env_profile:
        unity_profile = env_profile.strip() or None

    env_host = os.getenv("DATABRICKS_HOST")
    if env_host:
        unity_url = env_host.strip() or None

    env_workspace_token = os.getenv("DATABRICKS_TOKEN") or os.getenv("DC43_DATABRICKS_TOKEN")
    if env_workspace_token:
        unity_token = env_workspace_token.strip() or None

    env_unity_sql_dsn = os.getenv("DC43_UNITY_CATALOG_SQL_DSN")
    if env_unity_sql_dsn:
        unity_sql_dsn = env_unity_sql_dsn.strip() or None

    env_unity_tags_enabled = os.getenv("DC43_UNITY_CATALOG_TAGS_ENABLED")
    if env_unity_tags_enabled is not None:
        unity_tags_enabled = _parse_bool(env_unity_tags_enabled, unity_tags_enabled)

    env_unity_tags_sql_dsn = os.getenv("DC43_UNITY_CATALOG_TAGS_SQL_DSN")
    if env_unity_tags_sql_dsn:
        unity_tags_sql_dsn = env_unity_tags_sql_dsn.strip() or None

    env_link_builders = os.getenv("DC43_GOVERNANCE_LINK_BUILDERS")
    if env_link_builders:
        for chunk in env_link_builders.split(","):
            text = chunk.strip()
            if text:
                link_builder_specs.append(text)

    env_gov_store_type = os.getenv("DC43_GOVERNANCE_STORE_TYPE")
    if env_gov_store_type:
        gov_store_type = env_gov_store_type.strip().lower() or gov_store_type

    env_gov_root = os.getenv("DC43_GOVERNANCE_STORE")
    if env_gov_root:
        gov_root_value = _coerce_path(env_gov_root)
        gov_base_path_value = gov_root_value if gov_base_path_value is None else gov_base_path_value

    env_gov_base_path = os.getenv("DC43_GOVERNANCE_STORE_BASE_PATH")
    if env_gov_base_path:
        gov_base_path_value = _coerce_path(env_gov_base_path)

    env_gov_table = os.getenv("DC43_GOVERNANCE_STORE_TABLE")
    if env_gov_table:
        gov_table_value = env_gov_table.strip() or gov_table_value

    env_gov_status_table = os.getenv("DC43_GOVERNANCE_STATUS_TABLE")
    if env_gov_status_table:
        gov_status_table_value = env_gov_status_table.strip() or gov_status_table_value

    env_gov_activity_table = os.getenv("DC43_GOVERNANCE_ACTIVITY_TABLE")
    if env_gov_activity_table:
        gov_activity_table_value = env_gov_activity_table.strip() or gov_activity_table_value

    env_gov_link_table = os.getenv("DC43_GOVERNANCE_LINK_TABLE")
    if env_gov_link_table:
        gov_link_table_value = env_gov_link_table.strip() or gov_link_table_value

    env_gov_metrics_table = os.getenv("DC43_GOVERNANCE_METRICS_TABLE")
    if env_gov_metrics_table:
        gov_metrics_table_value = env_gov_metrics_table.strip() or gov_metrics_table_value

    env_gov_dsn = os.getenv("DC43_GOVERNANCE_STORE_DSN")
    if env_gov_dsn:
        gov_dsn_value = env_gov_dsn.strip() or gov_dsn_value

    env_gov_schema = os.getenv("DC43_GOVERNANCE_STORE_SCHEMA")
    if env_gov_schema:
        gov_schema_value = env_gov_schema.strip() or gov_schema_value

    env_gov_log_sql = os.getenv("DC43_GOVERNANCE_STORE_LOG_SQL")
    if env_gov_log_sql is not None:
        gov_log_sql = _parse_bool(env_gov_log_sql, gov_log_sql)

    env_gov_url = os.getenv("DC43_GOVERNANCE_STORE_URL")
    if env_gov_url:
        gov_base_url_value = env_gov_url.strip() or gov_base_url_value

    env_gov_token = os.getenv("DC43_GOVERNANCE_STORE_TOKEN")
    if env_gov_token:
        gov_token_value = env_gov_token.strip() or gov_token_value

    env_gov_token_header = os.getenv("DC43_GOVERNANCE_STORE_TOKEN_HEADER")
    if env_gov_token_header is not None:
        gov_token_header_value = env_gov_token_header.strip()

    env_gov_token_scheme = os.getenv("DC43_GOVERNANCE_STORE_TOKEN_SCHEME")
    if env_gov_token_scheme is not None:
        gov_token_scheme_value = env_gov_token_scheme.strip()

    env_gov_timeout = os.getenv("DC43_GOVERNANCE_STORE_TIMEOUT")
    if env_gov_timeout:
        gov_timeout_value = _coerce_float(env_gov_timeout, gov_timeout_value)

    # Preserve configuration order while dropping duplicates that may arrive via
    # the configuration file and environment variables.
    seen_builders: set[str] = set()
    ordered_builders: list[str] = []
    for spec in link_builder_specs:
        if spec in seen_builders:
            continue
        seen_builders.add(spec)
        ordered_builders.append(spec)

    return ServiceBackendsConfig(
        contract_store=ContractStoreConfig(
            type=store_type,
            root=root_value,
            base_path=base_path_value,
            table=table_value,
            base_url=base_url_value,
            dsn=dsn_value,
            schema=schema_value,
            token=store_token_value,
            timeout=timeout_value,
            contracts_endpoint_template=endpoint_template,
            default_status=default_status,
            status_filter=status_filter,
            catalog=catalog_value,
            log_sql=store_log_sql,
        ),
        data_product_store=DataProductStoreConfig(
            type=dp_type,
            root=dp_root_value,
            base_path=dp_base_path_value,
            table=dp_table_value,
            dsn=dp_dsn_value,
            schema=dp_schema_value,
            base_url=dp_base_url_value,
            catalog=dp_catalog_value,
            log_sql=dp_log_sql,
        ),
        data_quality=DataQualityBackendConfig(
            type=dq_type,
            base_url=dq_base_url_value,
            token=dq_token_value,
            token_header=dq_token_header_value,
            token_scheme=dq_token_scheme_value,
            headers=dq_headers_value,
            default_engine=dq_default_engine_value,
            engines=dq_engines_value,
        ),
        auth=AuthConfig(token=auth_token_value),
        unity_catalog=UnityCatalogConfig(
            enabled=unity_enabled,
            dataset_prefix=unity_prefix,
            workspace_profile=unity_profile,
            workspace_url=unity_url,
            workspace_token=unity_token,
            sql_dsn=unity_sql_dsn,
            tags_enabled=unity_tags_enabled,
            tags_sql_dsn=unity_tags_sql_dsn,
            static_properties=unity_static,
            static_tags=unity_static_tags,
        ),
        governance=GovernanceConfig(
            dataset_contract_link_builders=tuple(ordered_builders),
        ),
        governance_store=GovernanceStoreConfig(
            type=gov_store_type,
            root=gov_root_value,
            base_path=gov_base_path_value,
            table=gov_table_value,
            status_table=gov_status_table_value,
            activity_table=gov_activity_table_value,
            link_table=gov_link_table_value,
            metrics_table=gov_metrics_table_value,
            dsn=gov_dsn_value,
            schema=gov_schema_value,
            base_url=gov_base_url_value,
            token=gov_token_value,
            token_header=gov_token_header_value,
            token_scheme=gov_token_scheme_value,
            timeout=gov_timeout_value,
            headers=gov_headers_value,
            log_sql=gov_log_sql,
        ),
    )


def _stringify_path(path: Path | None) -> str | None:
    if path is None:
        return None
    try:
        return str(path)
    except Exception:  # pragma: no cover - defensive, should not happen
        return str(path)


def _contract_store_mapping(config: ContractStoreConfig) -> dict[str, Any]:
    mapping: dict[str, Any] = {}
    if config.type:
        mapping["type"] = config.type
    if config.root:
        mapping["root"] = _stringify_path(config.root)
    if config.base_path:
        mapping["base_path"] = _stringify_path(config.base_path)
    if config.table:
        mapping["table"] = config.table
    if config.base_url:
        mapping["base_url"] = config.base_url
    if config.dsn:
        mapping["dsn"] = config.dsn
    if config.schema:
        mapping["schema"] = config.schema
    if config.token:
        mapping["token"] = config.token
    if config.timeout != 10.0:
        mapping["timeout"] = config.timeout
    if config.contracts_endpoint_template:
        mapping["contracts_endpoint_template"] = config.contracts_endpoint_template
    if config.default_status and config.default_status != "Draft":
        mapping["default_status"] = config.default_status
    if config.status_filter:
        mapping["status_filter"] = config.status_filter
    if config.catalog:
        catalog_mapping: dict[str, Any] = {}
        for key, value in config.catalog.items():
            if not isinstance(value, tuple) or len(value) != 2:
                continue
            catalog_mapping[key] = {
                "data_product": value[0],
                "port": value[1],
            }
        if catalog_mapping:
            mapping["catalog"] = catalog_mapping
    if config.log_sql:
        mapping["log_sql"] = True
    return mapping


def _data_product_store_mapping(config: DataProductStoreConfig) -> dict[str, Any]:
    mapping: dict[str, Any] = {}
    if config.type:
        mapping["type"] = config.type
    if config.root:
        mapping["root"] = _stringify_path(config.root)
    if config.base_path:
        mapping["base_path"] = _stringify_path(config.base_path)
    if config.table:
        mapping["table"] = config.table
    if config.dsn:
        mapping["dsn"] = config.dsn
    if config.schema:
        mapping["schema"] = config.schema
    if config.base_url:
        mapping["base_url"] = config.base_url
    if config.catalog:
        mapping["catalog"] = config.catalog
    if config.log_sql:
        mapping["log_sql"] = True
    return mapping


def _data_quality_backend_mapping(config: DataQualityBackendConfig) -> dict[str, Any]:
    mapping: dict[str, Any] = {}
    if config.type:
        mapping["type"] = config.type
    if config.base_url:
        mapping["base_url"] = config.base_url
    if config.token:
        mapping["token"] = config.token
    if config.token_header and config.token_header != "Authorization":
        mapping["token_header"] = config.token_header
    if config.token_scheme and config.token_scheme != "Bearer":
        mapping["token_scheme"] = config.token_scheme
    if config.headers:
        mapping["headers"] = dict(config.headers)
    if config.default_engine and config.default_engine != "native":
        mapping["default_engine"] = config.default_engine
    if config.engines:
        mapping["engines"] = {name: dict(values) for name, values in config.engines.items()}
    return mapping


def _auth_mapping(config: AuthConfig) -> dict[str, Any]:
    if not config.token:
        return {}
    return {"token": config.token}


def _unity_catalog_mapping(config: UnityCatalogConfig) -> dict[str, Any]:
    mapping: dict[str, Any] = {}
    if config.enabled:
        mapping["enabled"] = True
    if config.dataset_prefix:
        mapping["dataset_prefix"] = config.dataset_prefix
    if config.workspace_profile:
        mapping["workspace_profile"] = config.workspace_profile
    if config.workspace_url:
        mapping["workspace_url"] = config.workspace_url
    if config.workspace_token:
        mapping["workspace_token"] = config.workspace_token
    if config.sql_dsn:
        mapping["sql_dsn"] = config.sql_dsn
    if config.tags_enabled:
        mapping["tags_enabled"] = True
    if config.tags_sql_dsn:
        mapping["tags_sql_dsn"] = config.tags_sql_dsn
    if config.static_properties:
        mapping["static_properties"] = dict(config.static_properties)
    if config.static_tags:
        mapping["static_tags"] = dict(config.static_tags)
    return mapping


def _governance_mapping(config: GovernanceConfig) -> dict[str, Any]:
    if not config.dataset_contract_link_builders:
        return {}
    return {
        "dataset_contract_link_builders": list(config.dataset_contract_link_builders)
    }


def _governance_store_mapping(config: GovernanceStoreConfig) -> dict[str, Any]:
    mapping: dict[str, Any] = {}
    if config.type:
        mapping["type"] = config.type
    if config.root:
        mapping["root"] = _stringify_path(config.root)
    if config.base_path:
        mapping["base_path"] = _stringify_path(config.base_path)
    if config.table:
        mapping["table"] = config.table
    if config.status_table:
        mapping["status_table"] = config.status_table
    if config.activity_table:
        mapping["activity_table"] = config.activity_table
    if config.link_table:
        mapping["link_table"] = config.link_table
    if config.metrics_table:
        mapping["metrics_table"] = config.metrics_table
    if config.dsn:
        mapping["dsn"] = config.dsn
    if config.schema:
        mapping["schema"] = config.schema
    if config.base_url:
        mapping["base_url"] = config.base_url
    if config.token:
        mapping["token"] = config.token
    if config.token_header and config.token_header != "Authorization":
        mapping["token_header"] = config.token_header
    if config.token_scheme and config.token_scheme != "Bearer":
        mapping["token_scheme"] = config.token_scheme
    if config.timeout != 10.0:
        mapping["timeout"] = config.timeout
    if config.headers:
        mapping["headers"] = dict(config.headers)
    if config.log_sql:
        mapping["log_sql"] = True
    return mapping


def config_to_mapping(config: ServiceBackendsConfig) -> dict[str, Any]:
    """Return a serialisable mapping derived from ``config``."""

    payload: dict[str, Any] = {}

    contract_mapping = _contract_store_mapping(config.contract_store)
    if contract_mapping:
        payload["contract_store"] = contract_mapping

    product_mapping = _data_product_store_mapping(config.data_product_store)
    if product_mapping:
        payload["data_product"] = product_mapping

    dq_mapping = _data_quality_backend_mapping(config.data_quality)
    if dq_mapping:
        payload["data_quality"] = dq_mapping

    auth_mapping = _auth_mapping(config.auth)
    if auth_mapping:
        payload["auth"] = auth_mapping

    unity_mapping = _unity_catalog_mapping(config.unity_catalog)
    if unity_mapping:
        payload["unity_catalog"] = unity_mapping

    governance_mapping = _governance_mapping(config.governance)
    if governance_mapping:
        payload["governance"] = governance_mapping

    governance_store_mapping = _governance_store_mapping(config.governance_store)
    if governance_store_mapping:
        payload["governance_store"] = governance_store_mapping

    return payload


def _toml_ready_value(value: Any) -> Any:
    """Return ``value`` converted into TOML-friendly primitives."""

    if isinstance(value, Path):
        return str(value)
    if isinstance(value, Mapping):
        return {str(key): _toml_ready_value(item) for key, item in value.items()}
    if isinstance(value, set):
        return [_toml_ready_value(item) for item in sorted(value, key=repr)]
    if isinstance(value, (list, tuple)):
        return [_toml_ready_value(item) for item in value]
    return value


def dumps(config: ServiceBackendsConfig) -> str:
    """Return a TOML string representation of ``config``."""

    mapping = config_to_mapping(config)
    if not mapping:
        return ""
    prepared = _toml_ready_value(mapping)
    if not prepared:
        return ""
    return _toml_dumps(prepared)


def dump(path: str | os.PathLike[str], config: ServiceBackendsConfig) -> None:
    """Write ``config`` to ``path`` in TOML format."""

    output = dumps(config)
    Path(path).write_text(output, encoding="utf-8")



from __future__ import annotations

import sys
from pathlib import Path

import pytest

from types import SimpleNamespace

from dc43_service_backends.bootstrap import (
    build_backends,
    build_contract_store,
    build_data_product_backend,
    build_data_quality_backend,
    build_governance_store,
)
from dc43_service_backends.config import (
    ContractStoreConfig,
    DataProductStoreConfig,
    DataQualityBackendConfig,
    GovernanceStoreConfig,
    ServiceBackendsConfig,
)
from dc43_service_backends.contracts.backend.stores import DeltaContractStore, FSContractStore
from dc43_service_backends.data_products import DeltaDataProductServiceBackend, LocalDataProductServiceBackend
from dc43_service_backends.data_quality.backend import (
    LocalDataQualityServiceBackend,
    RemoteDataQualityServiceBackend,
)
from dc43_service_backends.governance.backend import LocalGovernanceServiceBackend
from dc43_service_backends.governance.storage import InMemoryGovernanceStore
from dc43_service_backends.governance.storage.filesystem import FilesystemGovernanceStore


class _StubSparkResult:
    def collect(self) -> list[tuple]:
        return []

    def head(self, _: int) -> list[tuple]:
        return []


class _StubSparkSession:
    def __init__(self) -> None:
        self.calls: list[str] = []

    def sql(self, statement: str) -> _StubSparkResult:
        self.calls.append(statement)
        return _StubSparkResult()


class _StubSparkBuilder:
    def getOrCreate(self) -> _StubSparkSession:
        return _StubSparkSession()


class _StubSpark:
    builder = _StubSparkBuilder()


def _capture_create_engine(monkeypatch: pytest.MonkeyPatch) -> list[dict[str, object]]:
    sqlalchemy = pytest.importorskip("sqlalchemy")
    original = sqlalchemy.create_engine
    calls: list[dict[str, object]] = []

    def fake_create_engine(*args: object, **kwargs: object):
        calls.append({"args": args, "kwargs": dict(kwargs)})
        return original(*args, **kwargs)

    monkeypatch.setattr(sqlalchemy, "create_engine", fake_create_engine)
    return calls


def test_build_contract_store_filesystem(tmp_path: Path) -> None:
    cfg = ContractStoreConfig(type="filesystem", root=tmp_path)
    store = build_contract_store(cfg)
    assert isinstance(store, FSContractStore)


def test_build_contract_store_delta_log_sql(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr("dc43_service_backends.bootstrap.SparkSession", _StubSpark)
    cfg = ContractStoreConfig(type="delta", table="unity.contracts", log_sql=True)

    store = build_contract_store(cfg)

    assert isinstance(store, DeltaContractStore)
    assert getattr(store, "_log_sql", False) is True


def test_build_contract_store_sql(tmp_path: Path) -> None:
    pytest.importorskip("sqlalchemy")
    dsn = f"sqlite:///{tmp_path / 'contracts.db'}"
    cfg = ContractStoreConfig(type="sql", dsn=dsn, table="contracts")

    store = build_contract_store(cfg)

    from dc43_service_backends.contracts.backend.stores.sql import SQLContractStore

    assert isinstance(store, SQLContractStore)


def test_build_data_product_backend_memory() -> None:
    backend = build_data_product_backend(DataProductStoreConfig(type="memory"))
    assert isinstance(backend, LocalDataProductServiceBackend)


def test_build_data_product_backend_delta_log_sql(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr("dc43_service_backends.bootstrap.SparkSession", _StubSpark)
    backend = build_data_product_backend(
        DataProductStoreConfig(type="delta", table="unity.products", log_sql=True)
    )

    assert isinstance(backend, DeltaDataProductServiceBackend)
    assert getattr(backend._store, "_log_sql", False) is True  # type: ignore[attr-defined]


def test_build_backends_delta(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr("dc43_service_backends.bootstrap.SparkSession", _StubSpark)
    config = ServiceBackendsConfig(
        contract_store=ContractStoreConfig(type="delta", table="unity.contracts"),
        data_product_store=DataProductStoreConfig(type="delta", table="unity.products"),
    )

    suite = build_backends(config)
    assert isinstance(suite.contract._store, DeltaContractStore)
    assert isinstance(suite.data_product, DeltaDataProductServiceBackend)
    assert isinstance(suite.data_quality, LocalDataQualityServiceBackend)
    assert isinstance(suite.governance_store, InMemoryGovernanceStore)
    assert isinstance(suite.governance, LocalGovernanceServiceBackend)
    assert isinstance(suite.contract_store, DeltaContractStore)
    assert isinstance(suite.link_hooks, tuple)


def test_build_data_quality_backend_local() -> None:
    backend = build_data_quality_backend(DataQualityBackendConfig(type="local"))
    assert isinstance(backend, LocalDataQualityServiceBackend)


def test_build_data_quality_backend_http(monkeypatch: pytest.MonkeyPatch) -> None:
    called: dict[str, object] = {}

    class _StubClient:
        def __init__(self, **kwargs: object) -> None:
            called.update(kwargs)

        def evaluate(self, **_: object) -> object:  # pragma: no cover - exercised via interface
            return object()

        def describe_expectations(self, **_: object) -> list[dict[str, object]]:
            return []

    stub_module = SimpleNamespace(RemoteDataQualityServiceClient=_StubClient)
    monkeypatch.setitem(sys.modules, "dc43_service_clients.data_quality.client.remote", stub_module)

    backend = build_data_quality_backend(
        DataQualityBackendConfig(
            type="http",
            base_url="https://observability.local",
            token="secret",
            token_header="X-Token",
            token_scheme="",
            headers={"X-Correlation-ID": "abc"},
        )
    )

    assert isinstance(backend, RemoteDataQualityServiceBackend)
    assert called["base_url"] == "https://observability.local"
    assert called["token"] == "secret"
    assert called["token_header"] == "X-Token"
    assert called["token_scheme"] == ""
    assert called["headers"] == {"X-Correlation-ID": "abc"}


def test_build_governance_store_memory() -> None:
    store = build_governance_store(GovernanceStoreConfig(type="memory"))
    assert isinstance(store, InMemoryGovernanceStore)


def test_build_governance_store_filesystem(tmp_path: Path) -> None:
    cfg = GovernanceStoreConfig(type="filesystem", root=tmp_path)
    store = build_governance_store(cfg)
    assert isinstance(store, FilesystemGovernanceStore)
    assert (tmp_path / "status").exists()


def test_build_contract_store_sql_echo(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    calls = _capture_create_engine(monkeypatch)
    dsn = f"sqlite:///{tmp_path / 'contracts-log.db'}"
    cfg = ContractStoreConfig(type="sql", dsn=dsn, log_sql=True)

    build_contract_store(cfg)

    assert calls and calls[0]["kwargs"].get("echo") is True


def test_build_data_product_store_sql_echo(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    calls = _capture_create_engine(monkeypatch)
    dsn = f"sqlite:///{tmp_path / 'data-products-log.db'}"
    cfg = DataProductStoreConfig(type="sql", dsn=dsn, log_sql=True)

    backend = build_data_product_backend(cfg)

    assert calls and calls[0]["kwargs"].get("echo") is True
    assert isinstance(backend, LocalDataProductServiceBackend)


def test_build_governance_store_sql_echo(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    calls = _capture_create_engine(monkeypatch)
    dsn = f"sqlite:///{tmp_path / 'governance-log.db'}"
    cfg = GovernanceStoreConfig(type="sql", dsn=dsn, log_sql=True)

    store = build_governance_store(cfg)

    assert calls and calls[0]["kwargs"].get("echo") is True
    from dc43_service_backends.governance.backend.stores import SQLGovernanceStore

    assert isinstance(store, SQLGovernanceStore)


def test_build_governance_store_delta_with_dsn(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    calls = _capture_create_engine(monkeypatch)
    dsn = f"sqlite:///{tmp_path / 'governance-delta.db'}"
    cfg = GovernanceStoreConfig(type="delta", dsn=dsn, log_sql=True)

    store = build_governance_store(cfg)

    from dc43_service_backends.governance.backend.stores import SQLGovernanceStore

    assert isinstance(store, SQLGovernanceStore)
    assert calls and calls[0]["kwargs"].get("echo") is True
